# -*- coding: utf-8 -*-

import numpy as np
from .coverage import find_coverage
from .metagene import find_site_coverage
from .region_counts import find_region_counts
from ..settings import *
from ..io.file import flex_open 


################################################################

def quantify_experiment( 
        input_reads_file, 
        ref_names, 
        ref_lengths, 
        region_coordinates, 
        metagene_radius, 
        left_span, 
        right_span):
    """
    Does all essential quantification for an input bed file.
    The results are packed in a dictionary called "essential_profile".
    The input reads can come from any bed file. 
    Though in riboflow, when we call this function,
    the input reads are of a fixed length.

    Parameters
    ----------
    input_reads : str 
        In bed format. It can also be an open stream.

    ref_names   : list(str) 
        Reference names in an array.

    ref_lengths : list(int)
        Reference lengths with the same order as names.

    region_coordinates : list( [start_u5, stop_u5], 
                               [start_cd, stop_cd]
                               [start_u3, stop_u3])
                         
                         Annotation of the reference.
                         In triplets of the form [start, stop]
                         for UTR5, CDS and UTR3
    metagene_radius : int
                      nucleotides to the left / right of start / stop site
                      for metagene analysis

    left_span: int
               Nucleotides to the left of start / stop site for region counts

    left_span: int
               Nucleotides to the right of start / stop site for region counts


    Returns
    -------
    essential_profile : An array of 

    """

    with flex_open( input_reads_file, "r" ) as input_stream:
        coverage = find_coverage(  input_stream , ref_names, ref_lengths )

    essential_profile = dict()

    essential_profile[REF_DG_COVERAGE] = coverage    

    essential_profile[REF_DG_START_SITE_COV] = \
        find_site_coverage(coverage, metagene_radius, 
             region_coordinates, site_type = "start")

    essential_profile[REF_DG_STOP_SITE_COV] = \
        find_site_coverage(coverage, metagene_radius, 
             region_coordinates, site_type = "stop")

    essential_profile[REF_DG_REGION_COUNTS] = \
        find_region_counts( coverage, 
             region_coordinates, left_span, right_span )

    return essential_profile









# -*- coding: utf-8 -*-
import unittest
import os

import pandas as pd
import h5py

from test_data import *
import ribogadgets
from ribogadgets import create
from ribogadgets.settings import *
from ribogadgets.ribo import ribo

###########################################
REF_LEN_FILE = "ref_lengths_api.tsv"
TEST_RIBO = "test_api.ribo"

###########################################

# TO be completed
"""
class TestRiboAPI(unittest.TestCase):

    def setUp(self):
        with open(REF_LEN_FILE, "w") as output_stream:
            print(TRANSCRIPT_LENGTHS, file=output_stream)
        create.create_ribo_file(TEST_RIBO, "api_exp_1", REF_LEN_FILE)
        self.ribo = ribo(TEST_RIBO)
        
    def tearDown(self):
        os.remove(REF_LEN_FILE)
        del self.ribo
        os.remove(TEST_RIBO)

    def test_experiments(self):
        self.assertIn("api_exp_1", self.ribo.experiments)
        self.assertNotIn("sherlock", self.ribo.experiments)
"""

if __name__ == '__main__':
        
    unittest.main()


# -*- coding: utf-8 -*-
import os
##############################################################################

def separate_by_length(input_stream, 
                       length_max, length_min, file_prefix = ""):
    """
    Given a bed file handle and a length range [min length,  max length] 
    it partitions the file into bed files for each length in the given range
    file_prefix is pre_appended to each file name.
    So, for example, for length_min = 3, length_max=5, 
    3 files for lengths 3,4,5 are going to be created

    Parameters
    ----------
    input_stream : A stream coming from a bed file
    length_max   : maximum length (inclusive)
    length_min   : minimum length (inclusive)
    file_prefix  : Will be preappended to output file names

    Returns
    -------
    A list of separated file paths.

    """

    output_file_base = file_prefix + ".len_"

    assert length_min > 0
    assert length_max >= length_min

    file_handles = list()
    separated_files = list()

    for i in range(length_min, length_max + 1):
        this_file_path = output_file_base + str(i)
        separated_files.append(this_file_path)
        file_handles.append( open(this_file_path, "w") )  

    try:
        for this_line in input_stream:
            this_line = this_line.strip()
            line_contents = this_line.split()
            
            if len(line_contents) < 6:
                continue

            read_length = int( line_contents[2]) - int( line_contents[1] )
            if read_length < length_min or read_length > length_max:
                continue

            output_handle = file_handles[read_length - length_min]
            print(this_line, file=output_handle)

    except Exception as e:
        [ os.remove( output_file_base + str(i) ) \
             for i in range(length_min, length_max + 1)  ] 
        raise IOError("An error happened while separating reads by length.\n" + str(e))

    finally:
        input_stream.close()
        [ f.close() for f in file_handles ]

    return separated_files



# -*- coding: utf-8 -*-

from collections import OrderedDict
import numpy as np
import pandas as pd

from ..settings import *

###########################################################

def _merge_profile( profile_list ):
    array_size = len(profile_list[0])

    for a in profile_list[1:]:
        assert array_size == len(a)

    return np.concatenate( profile_list ) 

def _merge_region( region_counts, 
                   length_min, length_max ):
    pass


def _write_metagene( metagene_handle, 
                     start_site, stop_site, 
                     metagene_radius ):

    metagene_handle.create_dataset(
                            REF_DG_START_SITE_COV,
                            data        = start_site,
                            dtype       = SITE_COVERAGE_DT,
                            compression = DEFAULT_COMPRESSION, 
                            fletcher32  = DEFAULT_FLETCHER32)

    metagene_handle.create_dataset(
                            REF_DG_STOP_SITE_COV,
                            data        = stop_site,
                            dtype       = SITE_COVERAGE_DT,
                            compression = DEFAULT_COMPRESSION, 
                            fletcher32  = DEFAULT_FLETCHER32)

    metagene_handle.attrs[METAGENE_RADIUS_name] = metagene_radius


def _write_region_counts( region_handle, 
                          region_counts, 
                          left_span, right_span ):

    region_handle.create_dataset(
                            REF_DG_REGION_COUNTS,
                            data        = region_counts,
                            dtype       = DEFAULT_COUNT_DT,
                            compression = DEFAULT_COMPRESSION, 
                            fletcher32  = DEFAULT_FLETCHER32)

    region_handle.attrs[LEFT_SPAN_name]  = left_span
    region_handle.attrs[RIGHT_SPAN_name] = right_span


#####################################################################

def _write_coverage(exp_handle, coverages, store_coverage):
    """
    ??? Incomplete documentation
    """

    # First coverage is for the first length
    first_coverage    = np.concatenate( tuple( coverages[0].values() ) )
    # The length of first coverage gives us the total
    # number of nucleotides in this transcript set
    total_tcript_nucs = len(first_coverage)

    total_coverage_size = len(coverages) * len(first_coverage)

    # allocating space for the whole array first and thenm filling it up
    # should be more efficient than concatanating arrays one after the other
    # though we might need to optimize this part  
    coverage_np         = np.zeros( total_coverage_size, 
                                    dtype = TRANSCRIPT_COVERAGE_DT )

    coverage_np[0:total_tcript_nucs] = first_coverage

    x = total_tcript_nucs
    y = x + total_tcript_nucs

    for c in coverages[1:]:
        this_coverage     = np.concatenate( tuple( c.values() ) )
        coverage_np[x:y]  = this_coverage
        x  = y
        y += total_tcript_nucs 

    if store_coverage:
        exp_handle.create_group(REF_DG_COVERAGE)
        coverage_handle = exp_handle[REF_DG_COVERAGE] 

        coverage_handle.create_dataset(
                            REF_DG_COVERAGE,
                            data        = coverage_np, 
                            dtype       = TRANSCRIPT_COVERAGE_DT,
                            compression = DEFAULT_COMPRESSION, 
                            fletcher32  = DEFAULT_FLETCHER32)

    return coverage_np.sum()



#####################################################################

def write_profile( exp_handle , profiles,
                   length_min , length_max,
                   left_span,   right_span, 
                   metagene_radius,
                   store_coverage = False ):
    
    start_site_profiles = list( map( lambda x: x[REF_DG_START_SITE_COV], 
                                     profiles ) )
    stop_site_profiles  = list( map( lambda x: x[REF_DG_STOP_SITE_COV], 
                                     profiles ) )
    region_counts       = list( map( lambda x: x[REF_DG_REGION_COUNTS], 
                                     profiles ) )
    coverages           = list( map( lambda x: x[REF_DG_COVERAGE],
                                     profiles ))

    merged_start_site_profiles = _merge_profile( start_site_profiles )
    merged_stop_site_profiles  = _merge_profile( stop_site_profiles )
    merged_region_counts       = _merge_profile(region_counts)

    exp_handle.create_group(METAGENE_name)
    metagene_handle = exp_handle[METAGENE_name]
    _write_metagene( metagene_handle, 
                     merged_start_site_profiles, 
                     merged_stop_site_profiles,
                     metagene_radius )

    exp_handle.create_group(REF_DG_REGION_COUNTS)

    region_handle = exp_handle[REF_DG_REGION_COUNTS]

    _write_region_counts( region_handle, 
                          merged_region_counts,
                          left_span, right_span )

    total_reads = _write_coverage(exp_handle, coverages, 
                                  store_coverage)

    exp_handle.attrs[ATTRS_TOTAL_READS] = total_reads

    
    

        
    



# -*- coding: utf-8 -*-
import unittest

import os
from io import StringIO, BytesIO

import numpy as np
import h5py


from ribogadgets import create
from ribogadgets.core.coverage import find_coverage
from ribogadgets.core.get_gadgets import get_reference_names,\
                                         get_reference_lengths,\
                                         get_region_boundaries
from ribogadgets.settings import *
from ribogadgets.dump import *

from multilength_test_data import *
from ribogadgets.core import create_experiment

###########################################

NPROCESS = 4

class TestCreate(unittest.TestCase):

    def setUp(self):
        self.ref_len_file    = StringIO(TRANSCRIPT_LENGTHS)
        self.annotation_file = StringIO(TRANSCRIPT_ANNOTATION)
        self.alignment_file  = StringIO(READ_SET_1)

        self.handle = h5py.File(BytesIO())

        create.create_ribo(     self.handle, "merzifon", 
                                alignment_file  = self.alignment_file,
                                reference_name  = "appris_human_v2",
                                length_min      = 2,
                                length_max      = 5,
                                metagene_radius = METAGENE_RADIUS,
                                left_span       = LEFT_SPAN, 
                                right_span      = RIGHT_SPAN,
                                lengths_file    = self.ref_len_file,
                                store_coverage  = True,
                                annotation_file = self.annotation_file)

    def tearDown(self):
        self.handle.close()

    ### T E S T    S U M S  ################################    

    def test_dump_metagene_startsite(self):
        dumped_df = dump_metagene( 
                      ribo_handle     = self.handle, 
                      site_type       = REF_DG_START_SITE_COV, 
                      sum_lengths     = True, 
                      sum_references  = False,
                      range_lower     = 2, 
                      range_upper     = 4,
                      experiment_list = ["merzifon"] )["merzifon"]

        self.assertTrue( list(dumped_df.iloc[1]) == [0, 9, 6, 3, 0])

    def test_dump_metagene_stopsite(self):
        dumped_df = dump_metagene( 
                      ribo_handle     = self.handle, 
                      site_type       = REF_DG_STOP_SITE_COV, 
                      sum_lengths     = True, 
                      sum_references  = False,
                      range_lower     = 2, range_upper = 4,
                      experiment_list = ["merzifon"] )["merzifon"]

        # Start site is different from stop site
        self.assertTrue( not list(dumped_df.iloc[1]) == [0, 9, 6, 3, 0])

        self.assertTrue( list(dumped_df.iloc[2]) == [0, 0, 6, 0, 0])


    def test_dump_metagene_start_all_lengths(self):
        dumped_df = dump_metagene( 
                      ribo_handle     = self.handle, 
                      site_type       = REF_DG_START_SITE_COV, 
                      sum_lengths     = False, 
                      sum_references  = False,
                      range_lower     = 2, 
                      range_upper     = 3,
                      experiment_list = ["merzifon"] )["merzifon"]

        # Note that the first 3 is for the read length
        self.assertTrue( list(dumped_df.iloc[4] ) == [3,0,3,2,1,0] )


    def test_dump_metagene_start_all_lengths(self):
        dumped_df = dump_region(ribo_handle     = self.handle, 
                                region_name     = CDS_name,
                                sum_lengths     = True, 
                                sum_references  = False, 
                                range_lower     = 2 , 
                                range_upper     = 4,
                                experiment_list = [])["merzifon"]

        self.assertTrue( list(dumped_df[...]) ==  [11, 6 ,3])

    def test_dump_coverage(self):
        len_4_cov = dump_coverage( ribo_handle     = self.handle,
                                   experiment_name = "merzifon",
                                   range_lower     = 4,
                                   range_upper     = 4)


        self.assertEqual(len_4_cov["GAPDH"][7], 1)


        len_3_and_4_cov = dump_coverage( 
                                   ribo_handle     = self.handle,
                                   experiment_name = "merzifon",
                                   range_lower     = 3,
                                   range_upper     = 4)
        self.assertEqual(len_3_and_4_cov["VEGFA"][4], 5)

if __name__ == '__main__':
        
    unittest.main()


# -*- coding: utf-8 -*-
class RiboBaseError(Exception):
    pass

class AnnotationError(RiboBaseError):
	pass

class ExperimentDoesntExist(RiboBaseError):
    pass

class CoverageDoesntExist(RiboBaseError):
    pass

class InvalidName(RiboBaseError):
    pass

class MissingExperiment(RiboBaseError):
    pass

class InvalidLengthRange(RiboBaseError):
    pass


# -*- coding: utf-8 -*-
import unittest

import os
from io import StringIO, BytesIO

import numpy as np
import h5py


from ribogadgets import create
from ribogadgets.core.coverage import find_coverage
from ribogadgets.core.get_gadgets import get_reference_names,\
                                         get_reference_lengths,\
                                         get_region_boundaries
from ribogadgets.settings import *
from ribogadgets.metadata import *

from multilength_test_data import *
from ribogadgets.core import create_experiment

###########################################

NPROCESS = 4

class TestCreate(unittest.TestCase):

    def setUp(self):
        self.ref_len_file    = StringIO(TRANSCRIPT_LENGTHS)
        self.annotation_file = StringIO(TRANSCRIPT_ANNOTATION)
        self.alignment_file  = StringIO(READ_SET_1)

        self.handle = h5py.File(BytesIO())

        create.create_ribo(     self.handle, "merzifon", 
                                alignment_file  = self.alignment_file,
                                reference_name  = "appris_human_v2",
                                length_min      = 2,
                                length_max      = 5,
                                metagene_radius = METAGENE_RADIUS,
                                left_span       = LEFT_SPAN, 
                                right_span      = RIGHT_SPAN,
                                lengths_file    = self.ref_len_file,
                                store_coverage  = True,
                                annotation_file = self.annotation_file)


    def tearDown(self):
        self.handle.close()

    ### T E S T    S U M S  ################################

    def test_set_metadata(self):
        exp_handle  = self.handle[EXPERIMENTS_name]["merzifon"]
        ribo_handle = self.handle

        metadata_1  = \
            "cell-count : 5\nscientist : john doe\nsequencer : MiSeq"
        set_metadata(self.handle, "merzifon", metadata_1)

        self.assertTrue( "MiSeq" in exp_handle.attrs[USER_METADATA])

        metadata_2_str = \
            "cell-count : 10\nscientist : who doe\nsequencer : HiSeq-4000"
        metadata_2 = StringIO(metadata_2_str)
        set_metadata(self.handle, "merzifon", metadata_2)

        self.assertTrue( "MiSeq" not in exp_handle.attrs[USER_METADATA])

        self.assertTrue( "HiSeq-4000" in exp_handle.attrs[USER_METADATA])

        ribo_metadata = "pipelineversion : v1.0.2\ndate: 3-5-2018"

        set_metadata(self.handle, "", ribo_metadata)
        self.assertTrue("pipelineversion" in ribo_handle.attrs["metadata"])

        # Invalid Metadata

        invalid_metadata = "nonsens : 3\tsome-sense : 89"
        error = set_metadata(self.handle, "", invalid_metadata) 
        self.assertTrue("Invalid" in error)

    def test_get_metadata(self):
        exp_handle  = self.handle[EXPERIMENTS_name]["merzifon"]
        ribo_handle = self.handle

        metadata_1  = \
            "cell-count : 5\nscientist : john doe\nsequencer : MiSeq"
        set_metadata(self.handle, "merzifon", metadata_1)

        metadata_back_1 = get_metadata( self.handle, "merzifon" )
        self.assertEqual( str(metadata_back_1["cell-count"]), "5")
        self.assertEqual(metadata_back_1["sequencer"], "MiSeq")

        ribo_metadata   = "pipelineversion : v1.0.2\ndate: 3-5-2018"

        set_metadata(self.handle, "", ribo_metadata)
        ribo_metadata_back = get_metadata(self.handle, "")
        self.assertEqual(ribo_metadata_back["date"], "3-5-2018")

    def test_get_empty_metadata(self):
        exp_handle  = self.handle[EXPERIMENTS_name]["merzifon"]
        ribo_handle = self.handle

        metadata_back_1 = get_metadata( self.handle, "merzifon" )

        self.assertEqual(metadata_back_1, None)




    

if __name__ == '__main__':
        
    unittest.main()


from .main import *
from ..metadata import *

@cli.group()
def metadata():
    """
    Display, set or delete user-defined metadata

    If no name is given, the metadata of the ribo file is 
    set, displayed or deleted.
    """
    pass


@metadata.command()
@click.argument('ribo', type = click.Path( ))
@click.option('--name',  
               help     = "experiment name", 
               type     = click.STRING , 
               required = False)
@click.option('--meta',  
               help     = "Metadata File", 
               type     = click.Path(exists = True) , 
               required = True)
@click.option( '--force',
               is_flag = True,
               help    = 'Set metadata without prompting user.')
def set(ribo, name, meta, force):
    """Stores the metadata in the meta file.

    If no name is given, the metadata of the ribo file is set.

    If name is provided, the metadata of the corresponding experiment
    is set.

    The metadata must be in yaml format.

    \b
    Example Ribo Metadata File Contents:
          
        =================  ===================
        pipeline_name:     RiboFlow
        pipeline_version:  v1.0.2
        project:           elongation blocker
        =================  ===================


    \b
    Example Library Metadata File Contents:
        
       ===========   ========== 
        Cell_Line:   Human ESC
        Treatment:   Drug A   
        Enzyme:      RNASEI
       ===========   ==========      

    """

    set_metadata_wrapper(ribo_file  = ribo , 
                         name       = name , 
                         meta_file  = meta,
                         force      = force)


@metadata.command()
@click.argument('ribo', type = click.Path( ))
@click.option('--name',  
               help     = "experiment name", 
               type     = click.STRING , 
               required = False)
def get(ribo, name):
    """Displays user-defined metadata of the ribo file or experiment

    If no name is provided, metadata of the ribo file is displayed.

    If name is given, metadata of the corresponding experiment is displayed.
    """

    get_metadata_wrapper(ribo_file = ribo , 
                         name      = name)


@metadata.command()
@click.argument('ribo', type = click.Path( ))
@click.option('--name',  
               help     = "experiment name", 
               type     = click.STRING , 
               required = False)
@click.option( '--force',
               is_flag = True,
               help    = 'Set metadata without prompting user.')
def delete(ribo, name, force):
    """
    Deletes user-defined metadata of the ribo file or experiment

    If no name is provided, metadata of the ribo file is deleted.

    If name is given, metadata of the corresponding experiment is deleted.
    """

    delete_metadata_wrapper(ribo_file = ribo , 
                            name      = name,
                            force     = force)



# -*- coding: utf-8 -*-
import unittest
from unittest import mock
from unittest.mock import patch
import os
from io import StringIO, BytesIO

import numpy as np
import h5py

from ribogadgets import create
from ribogadgets.merge import merge_ribos
from ribogadgets.core.coverage import find_coverage
from ribogadgets.core.get_gadgets import get_reference_names,\
                                         get_reference_lengths,\
                                         get_region_boundaries,\
                                         get_experiment_names
from ribogadgets.settings import *
from ribogadgets.dump import *

from multilength_test_data import *
from ribogadgets.core import create_experiment

###########################################

INCOMPATIBLE_TRANSCRIPT_LENGTHS=\
"""GAPDH    20
VEGFA   22
MYC 17
EXTRA   56"""

incomp_len_file = "incompatible_lengths.tsv"

INCOMPATIBLE_TRANSCRIPT_ANNOTATION=\
"""GAPDH    0   5  UTR5    0   +
GAPDH   5  15    CDS 0   +
GAPDH   15  20    UTR3 0   +
VEGFA    0   4  UTR5    0   +
VEGFA   4  16    CDS 0   +
VEGFA   16  22    UTR3 0   +
MYC    0   3  UTR5    0   +
MYC   3  10    CDS 0   +
MYC   10  17    UTR3 0   +
EXTRA   0  20    UTR5 0   +
EXTRA   20  49    CDS 0   +
EXTRA   49  56    UTR3 0   +"""

incompatible_annotation = "incomp_annotation.bed"

###########################################

NPROCESS = 4

class TestCreate(unittest.TestCase):

    def setUp(self):
        self.tmp_files = list()


        self.ref_len_file       = StringIO(TRANSCRIPT_LENGTHS)
        self.annotation_file    = StringIO(TRANSCRIPT_ANNOTATION)
        self.alignment_file_1   = StringIO(READ_SET_1)
        self.alignment_file_2   = StringIO(READ_SET_2)
        self.in_ref_len_file    = StringIO(INCOMPATIBLE_TRANSCRIPT_LENGTHS)
        self.in_annotation_file = StringIO(INCOMPATIBLE_TRANSCRIPT_ANNOTATION)

        self.handle   = h5py.File(BytesIO() )
        self.handle_2 = h5py.File(BytesIO() )

        create.create_ribo(
                ribo            = self.handle, 
                experiment_name = "merzifon", 
                alignment_file  = self.alignment_file_1,
                reference_name  = "hg38",
                lengths_file    = self.ref_len_file, 
                annotation_file = self.annotation_file,
                metagene_radius = METAGENE_RADIUS, 
                left_span       = LEFT_SPAN, 
                right_span      = RIGHT_SPAN,
                length_min      = 2,
                length_max      = 5,  
                nprocess        = NPROCESS,
                tmp_file_prefix = "")
        self.ref_len_file.seek(0)
        self.annotation_file.seek(0)
        self.alignment_file_1 = StringIO(READ_SET_1)

        
        create.create_ribo(
                ribo            = self.handle_2, 
                experiment_name = "ankara", 
                alignment_file  = self.alignment_file_2,
                reference_name  = "hg38",
                lengths_file    = self.ref_len_file, 
                annotation_file = self.annotation_file,
                metagene_radius = METAGENE_RADIUS, 
                left_span       = LEFT_SPAN, 
                right_span      = RIGHT_SPAN,
                length_min      = 2,
                length_max      = 5,
                nprocess        = NPROCESS,
                tmp_file_prefix = "")
        self.ref_len_file.seek(0)
        self.annotation_file.seek(0)
        self.alignment_file_2 = StringIO(READ_SET_2)


        self.merged_ribo = h5py.File(BytesIO(), "w")
        merge_ribos( self.merged_ribo, [self.handle , self.handle_2] )
        

    def tearDown(self):
        self.handle.close()
        self.handle_2.close()
        self.merged_ribo.close()


    def test_merge_libnames(self):
        """
        Makes sure that the merged ribo file has the
        expected experiment names
        """

        self.handle.close()
        self.handle_2.close()
 
        merged_exp_names = set(get_experiment_names(self.merged_ribo)  )
        expected_libnames = set( ("ankara", "merzifon") )

        self.assertTrue( merged_exp_names == expected_libnames )

    
    def test_merge_regioncounts(self):
        cds_counts = dump_region(
                        ribo_handle     = self.merged_ribo, 
                        region_name     = CDS_name,
                        sum_lengths     = True, sum_references = False, 
                        range_lower     = 2, range_upper =5,
                        experiment_list = [])

        self.assertListEqual( list(cds_counts["ankara"]) , [5 , 3, 0])
        self.assertListEqual( list(cds_counts["merzifon"]) , [13 , 10, 3])

    def test_merge_start_site(self):

        start_site_cov_dict = dump_metagene( 
                                   ribo_handle     = self.merged_ribo, 
                                   site_type       = REF_DG_START_SITE_COV, 
                                   sum_lengths     = True, 
                                   sum_references  = True,
                                   range_lower     = 3, 
                                   range_upper     = 3,
                                   experiment_list = [] )

        self.assertListEqual(
             list(start_site_cov_dict["ankara"].iloc[0]), [0,0,0,0,0] )
             
        self.assertListEqual(
            list(start_site_cov_dict["merzifon"].iloc[0]), [3,5,3,2,0])

    def test_merge_single_ribo(self):
        single_merged = h5py.File (BytesIO() , "w" )

        merge_ribos( single_merged, [self.handle] )

        self.assertTrue( "merzifon" in get_experiment_names(single_merged) )


    def test_merge_incompatible_length(self):
        handle_incompatible = h5py.File(BytesIO() )
        ribo_out            = h5py.File(BytesIO() )

        create.create_ribo(
                ribo            = handle_incompatible, 
                experiment_name = "izmir", 
                alignment_file  = self.alignment_file_2,
                reference_name  = "hg38",
                lengths_file    = self.ref_len_file, 
                annotation_file = self.annotation_file,
                metagene_radius = METAGENE_RADIUS, 
                left_span       = LEFT_SPAN, 
                right_span      = RIGHT_SPAN,
                length_min      = 2,
                length_max      = 3, # incompatible parameter  
                nprocess        = NPROCESS,
                tmp_file_prefix = "")

        #merge_ribos( ribo_out, [self.handle, self.handle_2, handle_incompatible] )
        
        with self.assertRaises(ValueError):
            merge_ribos(ribo_out, 
                        [self.handle, self.handle_2, handle_incompatible] )


    def test_merge_incompatible_samelib(self):
        handle_incompatible = h5py.File(BytesIO() )
        ribo_out            = h5py.File(BytesIO() )

        create.create_ribo(
                ribo            = handle_incompatible, 
                experiment_name    = "ankara", # exception reason
                alignment_file  = self.alignment_file_2,
                reference_name  = "hg38",
                lengths_file    = self.ref_len_file, 
                annotation_file = self.annotation_file,
                metagene_radius = METAGENE_RADIUS, 
                left_span       = LEFT_SPAN, 
                right_span      = RIGHT_SPAN,
                length_min      = 2,
                length_max      = 5, 
                nprocess        = NPROCESS,
                tmp_file_prefix = "")

        #merge_ribos( ribo_out, [self.handle, self.handle_2, handle_incompatible] )
        
        with self.assertRaises(ValueError):
            merge_ribos( ribo_out, [self.handle, self.handle_2, 
                                    handle_incompatible] )
        

    
    def test_merge_incompatible_refname(self):
        handle_incompatible = h5py.File(BytesIO() )
        ribo_out            = h5py.File(BytesIO() )

        create.create_ribo(
                ribo            = handle_incompatible, 
                experiment_name = "izmir", 
                alignment_file  = self.alignment_file_2,
                reference_name  = "hg19", # incompatible reference name
                lengths_file    = self.ref_len_file, 
                annotation_file = self.annotation_file,
                metagene_radius = METAGENE_RADIUS, 
                left_span       = LEFT_SPAN, 
                right_span      = RIGHT_SPAN,
                length_min      = 2,
                length_max      = 5,  
                nprocess        = NPROCESS,
                tmp_file_prefix = "")

        
        with self.assertRaises(ValueError):
            merge_ribos(ribo_out, [self.handle, 
                                   self.handle_2, handle_incompatible] )


    def test_merge_incompatible_leftspan(self):
        handle_incompatible = h5py.File(BytesIO() )
        ribo_out            = h5py.File(BytesIO() )

        create.create_ribo(
                ribo            = handle_incompatible, 
                experiment_name = "izmir", 
                alignment_file  = self.alignment_file_2,
                reference_name  = "hg38",
                lengths_file    = self.ref_len_file, 
                annotation_file = self.annotation_file,
                metagene_radius = METAGENE_RADIUS, 
                left_span       = LEFT_SPAN + 3, # incomp span
                right_span      = RIGHT_SPAN,
                length_min      = 2,
                length_max      = 5,   
                nprocess        = NPROCESS,
                tmp_file_prefix = "")

        
        with self.assertRaises(ValueError):
            merge_ribos(ribo_out, [self.handle, 
                                   self.handle_2, handle_incompatible] )     



    def test_merge_incompatible_reference(self):
        handle_incompatible = h5py.File(BytesIO() )
        ribo_out            = h5py.File(BytesIO() )

        create.create_ribo(
                ribo            = handle_incompatible, 
                experiment_name = "izmir", 
                alignment_file  = self.alignment_file_2,
                reference_name  = "hg38",
                lengths_file    = self.in_ref_len_file, # incomp lens
                annotation_file = self.in_annotation_file, # incomp annot.
                metagene_radius = METAGENE_RADIUS, 
                left_span       = LEFT_SPAN,
                right_span      = RIGHT_SPAN,
                length_min      = 2,
                length_max      = 5,   
                nprocess        = NPROCESS,
                tmp_file_prefix = "")

        
        with self.assertRaises(ValueError):
            merge_ribos(ribo_out, [self.handle, 
                                   self.handle_2, handle_incompatible] )     



if __name__ == '__main__':
        
    unittest.main()


# -*- coding: utf-8 -*-

from .main import *
from ..merge import merge_ribo_files

@cli.command()
@click.argument(
    "out_ribo_path",
    nargs = 1,
    type  = click.Path(exists = False))
@click.argument(
    "in_ribo_paths",
    nargs = -1,
    type  = click.Path(exists = True))
def merge(out_ribo_path, in_ribo_paths):
    """
    Merges a set of given ribo files into one ribo file.
                     
    The input ribo files are merged into a new ribo file.
    The resulting ribo file has the union of the experiments
    of the input files

    The ribo files to be merged must be compatible:
    They must have the same\n 
    1) Reference Name \n
    2) Transcript Names & Transcript Lengths \n
    3) Annotation \n
    4) Ribo file parameters: \n
       a) Metagene Radius  \n
       b) Left Span & Right Span \n
       c) Min & Max Read Length \n
    5) They can not have overlapping experiment names

    This option is especially useful for Ribo-Seq Data
    processed together and needs to be analyzed together.
    """
    
    merge_ribo_files(out_ribo_path, in_ribo_paths)



# -*- coding: utf-8 -*-
import unittest
from unittest import mock
from unittest.mock import patch
import os
from io import StringIO, BytesIO

import pandas as pd
import h5py

from test_data import *
from ribogadgets import create
from ribogadgets.core.get_gadgets import *
from ribogadgets.settings import *

###########################################


class TestCreate(unittest.TestCase):

    def setUp(self):
        self.ref_len_file    = StringIO(TRANSCRIPT_LENGTHS)
        self.annotation_file = StringIO(TRANSCRIPT_ANNOTATION)
        self.alignment_file  = StringIO(READ_SET_1)
        self.handle          = h5py.File(BytesIO())

        create.create_ribo(
            self.handle, 
            experiment_name    = "merzifon", 
            alignment_file  = self.alignment_file,
            metagene_radius = 3,
            length_min      = 2, 
            length_max      = 5,
            left_span       = 3, 
            right_span      = 2,
            reference_name  = "appris_human_v2",
            lengths_file    = self.ref_len_file, 
            annotation_file = self.annotation_file)
        
    def tearDown(self):
        self.handle.close()


    def test_get_reference_names(self):   
        ref_names = get_reference_names(self.handle)
        self.assertListEqual( list(ref_names), ["GAPDH","VEGFA","MYC","BRCA"])

    def test_get_reference_lengths(self):
        ref_lengths = get_reference_lengths(self.handle)
        self.assertListEqual( list(ref_lengths), [1290, 875, 1462, 565])

    def test_get_number_of_references(self):
        number_of_refs = get_number_of_references(self.handle)
        self.assertEqual(number_of_refs, 4)

    def test_get_region_boundaries(self):
        all_boundaries = get_region_boundaries(self.handle)
        self.assertEqual( all_boundaries[1][0][1], 0 ) 
        self.assertEqual( all_boundaries[3][2][0], 561 )

    def test_get_experiment_names(self):
        self.handle[EXPERIMENTS_name].create_group("Antalya")
        libnames = get_experiment_names(self.handle)
        self.assertTrue( "Antalya" in libnames)
        self.assertTrue( "merzifon" in libnames)

    def test_get_read_length_range(self):
        len_lower , len_upper = get_read_length_range( self.handle )
        self.assertEqual(len_lower, 2)
        self.assertEqual(len_upper, 5)



if __name__ == '__main__':
        
    unittest.main()



# -*- coding: utf-8 -*-
import unittest
from unittest import mock
from unittest.mock import patch
import os
from io import StringIO, BytesIO

import numpy as np
import h5py

from test_data import *
from ribogadgets import create
from ribogadgets.core.coverage import find_coverage
from ribogadgets.core.metagene import find_site_coverage
from ribogadgets.core.get_gadgets import get_reference_names,\
                                         get_reference_lengths,\
                                         get_region_boundaries
from ribogadgets.settings import *



###########################################

TRANSCRIPT_LENGTHS=\
"""GAPDH    10
VEGFA   5
FLT 22
MYC 2"""


TRANSCRIPT_ANNOTATION=\
"""GAPDH    0   4  UTR5    0   -
GAPDH   4  7    CDS 0   -
GAPDH   7    10 UTR3   0   -
VEGFA   0   3 CDS 0   +
VEGFA   3 5 UTR3    0   +
FLT   0 6 UTR5    0   +
FLT   6 15 CDS    0   +
FLT   15 22 UTR3    0   +
MYC 0  1    CDS 0   -
MYC 1    2    UTR3    0   -"""

READ_SET_1=\
"""GAPDH 2  4  read_1 0   +
MYC 0  1  read_2 0 +
VEGFA 0  5  read_30 0 +
VEGFA 0  6  read_31 0 +
VEGFA 4  5  read_3 0 +
MYC 0  1  read_4 0 +
MYC 0  1  read_5 0 +
VEGFA 4  5  read_6 0 +
VEGFA 1  2  read_7 0 +
FLT 4  10  read_8 0 +
FLT 4  6  read_9 0 +
FLT 4  8  read_10 0 +
FLT 4  7  read_11 0 +
FLT 4  9  read_12 0 +
FLT 4  5  read_13 0 +
FLT 5  6  read_14 0 +
FLT 5  6  read_15 0 +
FLT 5  7  read_16 0 +
FLT 5  8  read_17 0 +
FLT 6  9  read_18 0 +
FLT 7  8  read_19 0 +
FLT 7  9  read_20 0 +
FLT 7  10  read_21 0 +
FLT 7  11  read_22 0 +
FLT 7  12  read_23 0 +
FLT 8  9  read_24 0 +
FLT 8  11  read_240 0 +
FLT 9  19  read_25 0 +
FLT 10  12  read_26 0 +
FLT 11  15  read_27 0 +
FLT 12  14  read_28 0 +
FLT 13  15  read_29 0 +
FLT 13  16  read_30 0 +
FLT 13  17  read_31 0 +
FLT 13  17  read_32 0 +
FLT 14  17  read_33 0 +
FLT 14  17  read_34 0 +
FLT 16  17  read_35 0 +
FLT 16  17  read_36 0 +
FLT 16  17  read_37 0 +
FLT 17  19  read_38 0 +"""

"""
Expected start site distribution, for radius=2:

GAPDH: [1,0,0,0,0]
VEGFA: [0, 0, 2, 1, 0]
FLT1: [6, 4, 1, 5, 2]
MYC: [0, 0, 3, 0, 0]
"""

"""
Expected stop site distribution, for radius = 2:

GAPDH: [0, 0, 0, 0, 0]
VEGFA: [1, 0, 0, 2, 0]
MYC: [0, 3, 0, 0, 0 ]
FLT1: [4, 2, 0, 3, 1]
"""

ACTUAL_START_SITE_COVERAGE = np.array( [[1, 0, 0, 0, 0], 
                                [0, 0, 2, 1, 0],
                                [6, 4, 1, 5, 2],
                                [0, 0, 3, 0, 0]] )

ACTUAL_STOP_SITE_COVERAGE = np.array( [ [0, 0, 0, 0, 0],
                                        [1, 0, 0, 2, 0],
                                        [4, 2, 0, 3, 1],
                                        [0, 3, 0, 0, 0]] )

###########################################
"""GAPDH    0   4  UTR5    0   -
GAPDH   4  7    CDS 0   -
GAPDH   7    10 UTR3   0   -
VEGFA   0   3 CDS 0   +
VEGFA   3 5 UTR3    0   +
FLT   0 6 UTR5    0   +
FLT   6 15 CDS    0   +
FLT   15 20 UTR3    0   +
MYC 0  1    CDS 0   -
MYC 1    1    UTR3    0   -"""


ANNOTATION = [\
               ([0, 4], [4, 7], [7, 10]),\
               ([0, 0], [0, 3], [3, 5]),\
               ([0, 6], [6, 15], [15, 20]),\
               ([0,0], [0, 1], [1,1])
              ]


def _get_transcripts(file_in_string):
    rows        = file_in_string.split("\n")
    pairs       = tuple( map( lambda x: x.split(), rows ) )
    ref_names   = tuple( map( lambda x: x[0], pairs ) )
    ref_lengths = tuple( map( lambda x: x[1], pairs ) )
    ref_lengths = tuple (map( int, ref_lengths ) )
    return (ref_names, ref_lengths)


class TestCreate(unittest.TestCase):

    def setUp(self):
        self.ref_len_file    = StringIO(TRANSCRIPT_LENGTHS)
        self.annotation_file = StringIO(TRANSCRIPT_ANNOTATION)
        self.alignment_file  = StringIO(READ_SET_1)
        self.handle          = h5py.File(BytesIO())
        
    def tearDown(self):
        self.handle.close()
        

    def test_start_site_coverage_solo(self):
        input_stream = StringIO(READ_SET_1)
        ref_names , ref_lengths = _get_transcripts(TRANSCRIPT_LENGTHS)

        coverage = find_coverage(input_stream,  
                                 ref_names , ref_lengths )

        site_coverage = find_site_coverage(coverage   = coverage, 
                                           radius     = 2, 
                                           annotation = ANNOTATION,
                                           site_type  = "start")

        comparison = (ACTUAL_START_SITE_COVERAGE == site_coverage)
        self.assertTrue( np.all( comparison ) )

    def test_stop_site_coverage_solo(self):
        input_stream = StringIO(READ_SET_1)
        ref_names , ref_lengths = _get_transcripts(TRANSCRIPT_LENGTHS)
        
        coverage = find_coverage(input_stream,  
                                 ref_names , ref_lengths )

        site_coverage = find_site_coverage(coverage   = coverage, 
                                           radius     = 2, 
                                           annotation = ANNOTATION,
                                           site_type  = "stop")

        comparison = (ACTUAL_STOP_SITE_COVERAGE == site_coverage)
        self.assertTrue( np.all( comparison ) )

    def test_start_site_coverage_from_ribo(self):
        create.initialize(self.handle, "merzifon", 
                          reference_name = "appris_human_v2")
        create.set_reference_names_and_lengths(
                                       self.handle, 
                                       self.ref_len_file)
        create.set_annotation( 
            h5_handle       = self.handle, 
            annotation_lines = TRANSCRIPT_ANNOTATION.split("\n") )

        ribo_annotation = get_region_boundaries(self.handle)
        ref_names       = get_reference_names(self.handle)
        ref_lengths     = get_reference_lengths(self.handle)
        input_stream    = StringIO(READ_SET_1)
        coverage        = find_coverage(input_stream,  
                                        ref_names , ref_lengths )

        site_coverage = find_site_coverage(coverage   = coverage, 
                                           radius     = 2, 
                                           annotation = ribo_annotation,
                                           site_type  = "start")

        comparison = (ACTUAL_START_SITE_COVERAGE == site_coverage)
        self.assertTrue( np.all( comparison ) )


if __name__ == '__main__':
        
    unittest.main()


# -*- coding: utf-8 -*-
from collections import OrderedDict
import time

import h5py
import pandas as pd
import numpy as np

from .settings import *
from .core.get_gadgets import *
from .core.metagene import *
from ._version import __format_version__, __version__

##########################################################




#####################################################################
#### C O M P A T I B I L I T Y    F U N C T I O N S
#####################################################################

def check_referenence_compatibility( first_pair, second_pair ):
    """
    ?? Incomplete documentation
    """

    reference_error = "Reference Error\n"

    first_reference_names  = get_reference_names( first_pair[0] )
    second_reference_names = get_reference_names( second_pair[0] )

    if len(first_reference_names) != len(second_reference_names):
        reference_error += "They have different number of references"
        reference_error += "{} != {}".format( len(first_reference_names),
                                              len(second_reference_names))
        return reference_error

    ref_name_comparison = first_reference_names == second_reference_names
    if not ref_name_comparison.all():
        return "Different reference (transcript) names."

    first_reference_lengths  = get_reference_lengths( first_pair[0] )
    second_reference_lengths = get_reference_lengths( second_pair[0] )

    ref_len_comparison = first_reference_lengths == second_reference_lengths
    if not ref_len_comparison.all():
        return "Different reference (transcript) lengths."

    return ""


def check_attribute_compatibility( first_pair, second_pair ):
    """
    ?? Incomplete documentation
    """

    attribute_error = "Attribute Error:\n"

    for attribute in ATTRS_ESSENTIAL_COMPATIBILITY:
        first_attr  = first_pair[0].attrs[attribute]
        second_attr = second_pair[0].attrs[attribute]

        if first_attr != second_attr:
            attribute_error += \
              " The {} atrribute is different:\n".format(attribute)
            attribute_error += "{first} != {second}".format(
                               first  = first_attr,
                               second = second_attr )
            return attribute_error

    return ""


############################################################################

def check_ribo_compatibility_pair( first_pair, second_pair ):
    """
    ?? Incomplete documentation
    """

    first_handle     = first_pair[0]
    first_identifier = first_pair[1]

    second_handle     = second_pair[0]
    second_identifier = second_pair[1]

    error_message = "The ribo files {}, {} are not compatible.\n".format(
                      first_identifier, second_identifier)

    # check attribute compatibility first
    attribute_error = \
        check_attribute_compatibility( first_pair, second_pair )
    if attribute_error:
        raise ValueError( error_message + attribute_error )

    ref_error = check_referenence_compatibility( first_pair, second_pair )
    if ref_error:
        raise ValueError( error_message + ref_error )


def check_if_common_libs_exist( ribo_handle_list ):
    """
    ?? Incomplete documentation
    """

    for i, ribo_1 in enumerate(ribo_handle_list):
        for ribo_2 in ribo_handle_list[i+1:]:
            experiments_1 = set(get_experiment_names(ribo_1[0]) )
            experiments_2 = set(get_experiment_names(ribo_2[0]) )

            common_experiments = experiments_1.intersection(experiments_2)

            if common_experiments :
                identifier_1  = ribo_1[1]
                identifier_2  = ribo_2[1]
                error_message = "The ribos {first} and {second} ".format(\
                                  first = identifier_1,
                                  second = identifier_2)
                error_message += " have common experiments:\n"
                error_message += str(common_experiments)
                raise ValueError(error_message)


def check_ribo_compatibility(ribo_handle_list):
    # Ribo_handle_list is a list of pairs where
    # an element of the list is of the form
    # (hdf5_file_handle, identifier)

    check_if_common_libs_exist(ribo_handle_list)

    for h in ribo_handle_list[1:]:
        check_ribo_compatibility_pair( ribo_handle_list[0], h )


#####################################################################
#### M A I N    F U N C T I O N S
#####################################################################

def initialize( h5_destination_handle, h5_source_handle ):
    """
    ?? Incomplete documentation
    """

    h5_source_handle.copy( REFERENCE_name,
               h5_destination_handle )
    h5_destination_handle.create_group(EXPERIMENTS_name)


###########################################################

def copy_metadata(h5_destination_handle, h5_source_handle):
    """
    ?? Incomplete documentation
    """

    for key in RIBO_METADATA_FOR_DISPLAY.keys():
        h5_destination_handle.attrs[key] = h5_source_handle.attrs[key]

def merge_ribos(destination_handle, source_list):
    """
    ?? Incomplete documentation
    """

    # If the sdource list is not coming from pairs,
    # then make it into pairs
    # this way, identifying incompatible ribo files or
    # handles is going to be easier
    if type(source_list[0]) not in (tuple, list):
        source_list = [ ( source_list[i], str(i) ) \
                         for i in range(len(source_list)) ]


    check_ribo_compatibility(source_list)

    source_handle_list = list( map(lambda x: x[0], source_list) )

    initialize(destination_handle, source_handle_list[0])
    copy_metadata(destination_handle, source_handle_list[0])

    for ribo_handle in source_handle_list:

        for lib in get_experiment_names(ribo_handle):
            exp_path = EXPERIMENTS_name + "/" + lib
            #destination_handle.create_group(exp_path)
            ribo_handle.copy( exp_path, destination_handle[EXPERIMENTS_name] )



def merge_ribo_files(destination_file, source_file_list):
    """
    Merges the experiments in the given source files and writes
    the result in the destination file.

    The ribo files need to be compatible
    (same left / right span, same metagene radius, same reference)

    Because of the compatibility, parameters (attributes), refrence
    etc. of the new file is the same as the merged file,

    The source files are not allowed to have experiments of the same name
    as this creates ambiguity.

    Parameters
    ----------
    destination_file : Destination ribo file path

    source_file_list : List of ribo file paths to be merged

    """

    if len(source_file_list) < 2:
        print("Please provide at least two input ribo files")
        exit(1)

    source_handle_list      = [ (h5py.File(f , "r"), f )
                                for f in source_file_list ]
    destination_ribo_handle = h5py.File( destination_file, "w" )

    merge_ribos( destination_ribo_handle, source_handle_list )

    [s[0].close() for s in source_handle_list]
    destination_ribo_handle.close()



# -*- coding: utf-8 -*-

import click

from .main import *

from ..plot import *
from ..settings import *



@cli.group()
def plot():
    """Generate some basic plots for ribo files."""
    pass


@plot.command()
@click.argument("ribo",      type=click.Path(exists=True))
@click.argument("experiments", nargs = -1,)
@click.option('-s', '--site',
              type     = click.Choice(['start', 'stop']),
              required = True,
              help     = 'Site type.')
@click.option('-o', '--out',
              required = True,
              help     = 'Output file.')
@click.option('--lowerlength',
              required = False ,
              type     = click.INT,
              help     = 'Lower read length')
@click.option('--upperlength',
              required = False ,
              type     = click.INT,
              help     ='Upper read length')
@click.option('-t', '--title',
              help     = 'Plot title.')
@click.option( '--normalize',
               is_flag = True,
               help    = 'Normalize by total metagene site coverage')
@click.option( '--dump', "-d",
               type    = click.Path(),
               help    = 'Dump the data to csv file')
def metagene(ribo, 
             site, 
             out, 
             experiments, 
             title,
             normalize,
             lowerlength, 
             upperlength, 
             dump ):
    """Generates metagene plots.
    
    The x-axis is the relative nucleotide positions
    and the start /stop site is at the center (origin at 0).
    The y-axis is the raw or normalized coverage.

    At most 7 experiments can be provided for a single plot.

    For a given start or stop site, the coverage around
    the site of interest is plotted. 

    The supported output extensions are pdf and png.

    If a length range is not provided, the minimum and the maximum
    ranges from the ribo file are read and used as the range.
    The values in the length range are aggregated to generate
    the metagene plot.

    \b
     
    Examples:  
     1) Get coverage data, of WT, for read length 21 in zipped bedgraph format

       .. code:: bash 

               ribog plot metagene -s start -o hela_1.pdf project.ribo HeLa_1


     2) Two experiments in one plot, stop site, with normalization

       .. code:: bash 

               ribog plot metagene -s start --normalize -o hela_1_2.pdf project.ribo HeLa_1 HeLa_2
               
     3) Plot start site usiong footprints of length 20,21,22
     Dump the data to out.csv file.

       .. code:: bash 

               ribog plot metagene -s start \\
                 --lowerlength 20 upperlength 22 \\
                 -o hela_1.pdf \\ 
                 -d out.csv \\
                 project.ribo HeLa_1             
    """

    return plot_metagene_wrapper( 
                           ribo_file        = ribo, 
                           output_file      = out,
                           site_type        = site, 
                           range_lower      = lowerlength, 
                           range_upper      = upperlength,
                           experiment_list  = experiments,
                           title            = title,
                           normalize        = normalize,
                           dump_to_file     = dump)




@plot.command()
@click.argument("ribo",      type = click.Path(exists=True))
@click.argument("experiments", nargs = -1,)
@click.option('-r', '--region',
              type     = click.Choice(EXTENDED_REGION_names),
              required = True,
              help     = 'Region type.')
@click.option('-o', '--out',
              required = True,
              help     = 'Output file in bed format')
@click.option('-t', '--title',
              help     = 'Plot title.')
@click.option( '--normalize',
               is_flag = True,
               help    ='Normalize by total metagene site coverage')
@click.option( '--dump', "-d",
               type    = click.Path(),
               help    = 'Dump the data to csv file')
def lengthdist(ribo, out, experiments, title,
                   normalize, region, dump ):
    """Plots the distribution of the ribosome footprint lengths.

    The x-axis is the length of the protected ribosome footprints.
    The y-axis is the raw or normalized frequecies.

    At most 7 experiments can be provided for a single plot.

    Pdf and png output formats are supported.
    If "dump" option is provided, the data is written 
    to the provided file path.

    If the frequencies are normalized using the "--normalize" option, 
    the y-axis becomes the percentages of the frequencies. 

    \b
     
    Examples:  
     1) Plot CDS length distribution of exp_1 and exp_2 
     and normalize the frequencies.

       .. code:: bash 

               ribog plot lengthdist  \\
                  -o multiple_dist.pdf \\
                  -r CDS --normalize \\
                  project.ribo exp_1 exp_2

     2)Plot only main_exp and write the data to out.csv.

       .. code:: bash 

               ribog plot lengthdist \\
                   -d out.csv \\
                   -o main_exp.pdf \\
                   -r CDS \\
                   project.ribo  main_exp

    """

    return plot_lengthdist_wrapper( 
                           ribo_file        = ribo, 
                           output_file      = out,
                           region_type      = region, 
                           experiment_list  = experiments,
                           title            = title,
                           normalize        = normalize,
                           dump_to_file     = dump)


@plot.command()
@click.argument("ribo",      type = click.Path(exists=True))
@click.argument("experiments", nargs = -1,)
@click.option('-o', '--out',
              required = True,
              help     = 'Output file')
@click.option('--lowerlength',
              required = False ,
              type     = click.INT,
              help     = 'Lower read length')
@click.option('--upperlength',
              required = False ,
              type     = click.INT,
              help     ='Upper read length')
@click.option('-t', '--title',
              help     = 'Plot title.')
@click.option( '--dump', "-d",
               type    = click.Path(),
               help    = 'Dump the data to csv file')
def regioncounts(ribo, out, experiments, 
                 upperlength, lowerlength, title, dump):
    """Generates barplots of the percentages of the UTR5, CDS and UTR3 counts.

    The raw counts are saved to a csv file if "--dump" option is provided.

    If a length range is not provided, then the range is determined
    using the mininum and the maximum read lengths in the ribo file.

    \b
     
    Examples:  
     1) Plot the region counts for the experiment names sample
     for lengths from 29 to 31

      .. code:: bash 

         ribog plot regioncounts --lowerlength 29 --upperlength 31 \\
             --out sample.region_counts.pdf test.ribo sample

     2) Plot region counts for the two experiments WT and DrugA
     for all lengths combined

      .. code:: bash 

         ribog plot regioncounts --out wt_anddrug.pdf other.ribo WT DrugA 
    """

    return plot_region_counts_wrapper(
                        ribo_file        = ribo, 
                        experiment_list  = experiments, 
                        range_lower      = lowerlength, 
                        range_upper      = upperlength, 
                        title            = title,
                        output_file      = out, 
                        dump_to_file     = dump)



# -*- coding: utf-8 -*-
import unittest
import os
from io import StringIO
import subprocess

import numpy as np
import h5py

from ribogadgets import create
from ribogadgets.core.coverage import find_coverage
from ribogadgets.core.get_gadgets import get_reference_names,\
                                         get_reference_lengths,\
                                         get_region_boundaries
from ribogadgets.settings import *
from ribogadgets.core.quantify import quantify_experiment

from multilength_test_data import *
from ribogadgets.core import create_experiment

###########################################

NPROCESS = 4

ribo_output_file = "cli_create.ribo"

EXPECTED_2_and_3_COVERAGE=\
"""GAPDH\t0\t1\t1
GAPDH\t1\t2\t2
GAPDH\t3\t4\t4
GAPDH\t4\t5\t4
GAPDH\t5\t6\t1
GAPDH\t6\t7\t2
GAPDH\t7\t8\t1
GAPDH\t8\t9\t2
GAPDH\t9\t10\t1
GAPDH\t10\t11\t2
GAPDH\t13\t14\t6
GAPDH\t14\t15\t3
GAPDH\t15\t16\t3
GAPDH\t16\t17\t1
GAPDH\t18\t19\t1
GAPDH\t19\t20\t1
VEGFA\t3\t4\t5
VEGFA\t4\t5\t3
VEGFA\t5\t6\t1
VEGFA\t10\t11\t3
VEGFA\t16\t17\t3
MYC\t0\t1\t1
MYC\t6\t7\t1
MYC\t10\t11\t3"""

EXPECTED_5_COVERAGE = \
"""GAPDH\t1\t2\t1
GAPDH\t3\t4\t1
GAPDH\t5\t6\t1
GAPDH\t8\t9\t1
GAPDH\t10\t11\t1
GAPDH\t14\t15\t1
VEGFA\t3\t4\t1
VEGFA\t10\t11\t4
VEGFA\t16\t17\t2
MYC\t0\t1\t1
MYC\t10\t11\t7"""

EXPECTED_2_and_3_VEGF_TSV_COVERAGE = \
"VEGFA\t0,0,0,5,3,1,0,0,0,0,3,0,0,0,0,0,3,0,0,0,0,0"


class TestCLIBase(unittest.TestCase):

    def run_command(self, command):
      """
      Command must be a list.
      See examples below,
      """

      command_str    = tuple( map( str, command ) )
      process        = subprocess.Popen(command_str, 
                                        stdout = subprocess.PIPE,
                                        stderr = subprocess.PIPE)
      output, error  = process.communicate()
      output_str     = output.decode()
      error_str      = error.decode()

      return (output_str, error_str) 



    def setUp(self):
        self.tmp_files = list()

        with open(REF_LEN_FILE, "w") as output_stream:
            print(TRANSCRIPT_LENGTHS, file=output_stream)
            self.tmp_files.append(REF_LEN_FILE)

        with open(ANNOTATION_FILE, "w") as output_stream:
            print(TRANSCRIPT_ANNOTATION, file=output_stream)
            self.tmp_files.append(ANNOTATION_FILE)

        with open(ALIGNMENT_FILE_1, "w") as output_stream:
            print(READ_SET_1, file= output_stream)
            self.tmp_files.append( ALIGNMENT_FILE_1 )

        create_command = ["ribog", "create", 
                          "-a", ALIGNMENT_FILE_1,
                          "--name",          "merzifon", 
                          "--reference",     "hg38",
                          "--lengths",        REF_LEN_FILE, 
                          "--annotation",     ANNOTATION_FILE,
                          "--metageneradius", METAGENE_RADIUS, 
                          "--leftspan",       LEFT_SPAN, 
                          "--rightspan",      RIGHT_SPAN,
                          "--lengthmin",      2, 
                          "--lengthmax",      5, 
                          "-n",               4,
                          ribo_output_file]

        output, error = self.run_command(create_command)
        print(output)
        self.tmp_files.append( ribo_output_file )


    def tearDown(self):
        [ os.remove(f) for f in self.tmp_files ]

#######################################################################  


class TestCLIDump(TestCLIBase):


    def test_dump_annotation(self):
        command_pieces  = ["ribog", "dump", "annotation", ribo_output_file]
        output, error   = self.run_command(command_pieces)
        output_lines    = output.split("\n")
        expected_line_2 = ["GAPDH", "5", "15", "CDS", "0", "+"]
        output_line2    = output_lines[1].split()
        self.assertTrue( expected_line_2 == output_line2 )


    def test_dump_ref_lengths(self):
        command_pieces     = ["ribog", "dump", "reference-lengths", 
                                ribo_output_file]
        output, error      = self.run_command(command_pieces)
        output_lines       = output.split("\n")
        expected_last_line = ["MYC", "17"]

        self.assertTrue( expected_last_line, output_lines[2] )

    def test_dump_region(self):
        #######################################
        # CDS, len 2 : 3 , no sum
        command_pieces = ["ribog", "dump", "region-counts", 
                          "--experiment",  "merzifon",
                          "--lowerlength", "2", 
                          "--upperlength", "3", 
                          "--region", "CDS", 
                          ribo_output_file]

        output, error = self.run_command(command_pieces)
        output_lines  = output.split("\n")

        expected_output_lines = [",{},merzifon".format(DF_READLENGTH),
                                 "GAPDH,2,2", "VEGFA,2,1", "MYC,2,0",
                                 "GAPDH,3,3", "VEGFA,3,2", "MYC,3,1"]

        expected_output = "\n".join(expected_output_lines)
        
        for x, y in zip( output_lines, expected_output_lines ):
            self.assertEqual(x, y)

        # CDS, len 2 : 3 , sum across transcripts
        command_pieces = ["ribog", "dump", "region-counts", 
                          "--experiment",  "merzifon",
                          "--lowerlength", "2", 
                          "--upperlength", "3",
                          "--sumtrans", 
                          "--region", "CDS",  
                          ribo_output_file]

        output, error       = self.run_command(command_pieces)
        output_str          = output.strip()
        expected_output_str = "read_length,merzifon\n2,3\n3,6"

        self.assertEqual(output_str, expected_output_str)

        # CDS, len 2 : 3 , sum across lengths
        command_pieces = ["ribog", "dump", "region-counts", 
                          "--experiment",  "merzifon",
                          "--lowerlength", "2", 
                          "--upperlength", "3",
                          "--sumlengths",
                          "--region", "CDS", 
                          ribo_output_file]

        output, error = self.run_command(command_pieces)
        output_str          = output.strip()
        expected_output_str = ",merzifon\nGAPDH,5\nVEGFA,3\nMYC,1"

        self.assertEqual(output_str, expected_output_str)

        # CDS, len 2 : 3 , sum across lengths and transcripts
        command_pieces = ["ribog", "dump", "region-counts", 
                          "--experiment",  "merzifon",
                          "--lowerlength", "2", 
                          "--upperlength", "3",
                          "--sumlengths",  "--sumtrans", 
                          "--region", "CDS",  
                          ribo_output_file]

        output, error = self.run_command(command_pieces)
        output_str          = output.strip()
        expected_output_str = ",0\nmerzifon,9"
        self.assertEqual(output_str, expected_output_str)


    def test_dump_metagene(self):
        
        #######################################
        # CDS, len 2 : 3 , sum transcripts
        command_pieces = ["ribog", "dump", "metagene", 
                          "--experiment",  "merzifon",
                          "--lowerlength", "2", 
                          "--upperlength", "3", 
                          "--site",        "start", 
                          ribo_output_file]

        output, error  = self.run_command(command_pieces)
        output_str     = output.strip() 
        expected_lines = [ ",".join( (DF_LIBRARY_NAME, DF_READLENGTH, 
                                        "-2", "-1", "0", "1", "2")),
                           ",".join(("merzifon", "2","1","4","1","1", "1")),
                           ",".join(("merzifon", "3","3","5","3","2","0")) ]
        expected_str = "\n".join(expected_lines)

        self.assertEqual(expected_str, output_str)    


        # CDS, len 2 : 3 , No sum
        command_pieces = ["ribog", "dump", "metagene", 
                          "--experiment",  "merzifon",
                          "--lowerlength", "2", 
                          "--upperlength", "3", 
                          "--nosumtrans",
                          "--site",        "start", 
                          ribo_output_file]

        output, error  = self.run_command(command_pieces)
        output_str     = output.strip() 
        output_lines   = output_str.split()

        self.assertEqual(output_lines[0], 
                 "experiment_name,{},read_length,-2,-1,0,1,2"\
                       .format(DF_TRANSCRIPT))
        self.assertEqual(output_lines[6], "merzifon,MYC,3,0,0,0,0,0")


        # CDS, len 2 : 3 , Sum lengths
        command_pieces = ["ribog", "dump", "metagene", 
                          "--experiment",  "merzifon",
                          "--lowerlength", "2", 
                          "--upperlength", "3", 
                          "--nosumtrans", 
                          "--sumlengths",
                          "--site", "start", 
                          ribo_output_file]

        output, error = self.run_command(command_pieces)
        output_str    = output.strip() 
        output_lines  = output_str.split()

        self.assertEqual("merzifon,VEGFA,0,5,3,1,0", output_lines[2])
        
        # CDS, len 2 : 3 , Sum lengths and transcripts on stop site

        command_pieces = ["ribog", "dump", "metagene", 
                          "--experiment",     "merzifon",
                          "--lowerlength", "2", 
                          "--upperlength", "3", 
                          "--sumlengths",
                          "--site", "stop", 
                          ribo_output_file]

        output, error  = self.run_command(command_pieces)
        output_str     = output.strip()
        expected_lines = output_str.split()

        self.assertEqual(expected_lines[1], "merzifon,6,3,9,1,0")

        # Run without a length range
        # So this should take length min oto length max by default
        command_pieces = ["ribog", "dump", "metagene", 
                          "--experiment", "merzifon", 
                          "--site",   "start", 
                          ribo_output_file]

        output, error  = self.run_command(command_pieces)
        output_str     = output.strip()

        expected_lines = [ ",".join( (DF_LIBRARY_NAME, DF_READLENGTH, 
                                        "-2", "-1", "0", "1", "2")),
                           ",".join(("merzifon", "2","1","4","1","1", "1")),
                           ",".join(("merzifon", "3","3","5","3","2","0")),
                           ",".join(("merzifon", "4","2","7","7","2","1")),
                           ",".join(("merzifon", "5","1","1","1","0","0")) ]
        expected_str   = "\n".join(expected_lines)

        self.assertEqual(expected_str, output_str) 


    def test_dump_coverage(self):
        command_pieces = ["ribog", "dump", "coverage",
                          "--lowerlength", "2", 
                          "--upperlength", "3",
                          "--format",      "bg",
                          ribo_output_file, "merzifon" ]

        output, error  = self.run_command(command_pieces)
        output_str     = output.strip()

        self.assertEqual(output_str, EXPECTED_2_and_3_COVERAGE)

        command_pieces = ["ribog", "dump", "coverage",
                          "--lowerlength", "5", 
                          "--upperlength", "5",
                          "--format",      "bg",
                          ribo_output_file, "merzifon" ]

        output, error  = self.run_command(command_pieces)
        output_str     = output.strip()

        self.assertEqual(output_str, EXPECTED_5_COVERAGE)





if __name__ == '__main__':
        
    unittest.main()



# -*- coding: utf-8 -*-

import h5py
import pandas as pd
import numpy as np

from ..settings import *
from .exceptions import *

###########################################################

def get_reference_names(h5_handle):
    """
    Returns reference (transcript) names as a string array 
    """
    return h5_handle[REFERENCE_name]\
                    [REF_DG_REFERENCE_NAMES][...].astype(str)


def get_reference_lengths(h5_handle):
    """
    Returns reference lengths in the same order as
    reference names (get_reference_names)
    """
    return h5_handle[REFERENCE_name]\
                    [REF_DG_REFERENCE_LENGTHS][...]

def get_number_of_references(h5_handle):
    """
    Returns the total number of transcripts (references)
    Use to create the ribo file
    """
    return len(h5_handle[REFERENCE_name]\
                        [REF_DG_REFERENCE_NAMES][...])

def get_region_boundaries(h5_handle):
    """
    Returns an array of pairs where pairs are of the form
    ( [UTR5_start=0, UTR5_end], [CDS_start, CDS_end], [UTR3_start, UTR3_end] )
    The coordinates are in bed format.
    Therefore it is zero based, first value is inclusive 
    and the second value is exclusive.
    """
    annotation_matrix = h5_handle[REFERENCE_name][REF_ANNOTATION_NAME][...]
    boundaries = map( lambda x: [(0, x[0]) , (x[0], x[1]), (x[1], x[2])] , 
                      annotation_matrix )

    return tuple(boundaries)

def get_experiment_names(ribo):
    """
    Returns experiment names in a ribo file.
    Initially, a ribo file has only one experiment
    But ribo files can be merged to contain more than one experiment.
    """
    experiments_handle = ribo[EXPERIMENTS_name]
    return tuple( experiments_handle.keys() )


def get_read_length_range(ribo):
    """
    Returns the minimum and maximum read lengths
    """
    return ( int(ribo.attrs[LENGTH_MIN_name]), 
             int(ribo.attrs[LENGTH_MAX_name]) )

def get_metagene_radius(ribo):
    """
    Returns the number of nucleotides to the left and right of
    start / stop sites in metagene analysis tables.
    """
    return ribo.attrs[ATTRS_METAGENE_RADIUS]

def get_left_right_spans(ribo):
    """
    In counting number of reads going to each region (UTR5, CDS and UTR3),
    We artificially introduce two regions

    UTR5_junction: A region between UTR5 and CDS containing the start site
                   [start_site - left_span, start_site + leftspan]

    UTR3_junction: A region between CDS and UTR3 containing the stop site
                   [stop_site - left_span, stop_site + leftspan]

    This function returns a tuple of the form:
    (left_span, right_span)

    """
    return (ribo.attrs[ATTRS_LEFT_SPAN],\
            ribo.attrs[ATTRS_RIGHT_SPAN])


def has_coverage_data(ribo, experiment):
    """
    Tells if the given experiment has coverage data
    stored.
    Note that storage of coverage data is optional so
    it may not exist in all experiments.
    """

    exp_handle = ribo[EXPERIMENTS_name][experiment]

    # Note that coverage exists  inside a datagroup
    # So we check the existince of the datagroup 
    # and the dataset 
    if REF_DG_COVERAGE not in exp_handle.keys() or \
        REF_DG_COVERAGE not in exp_handle[REF_DG_COVERAGE].keys():
        return False
    else:
        return True

def get_total_reads(ribo, experiment):

    exp_handle = ribo[EXPERIMENTS_name][experiment]

    return exp_handle.attrs[ATTRS_TOTAL_READS]

def rnaseq_exists(ribo, experiment):
    exp_handle = ribo[EXPERIMENTS_name][experiment]

    return  RNASEQ_name in exp_handle.get(RNASEQ_name, [])





# -*- coding: utf-8 -*-
import unittest
from unittest import mock
from unittest.mock import patch
import os
from io import StringIO
import subprocess
import shutil

import numpy as np
import h5py
import pipes

from ribogadgets import create
from ribogadgets.core.coverage import find_coverage
from ribogadgets.core.get_gadgets import get_reference_names,\
                                         get_reference_lengths,\
                                         get_region_boundaries
from ribogadgets.settings import *

from multilength_test_data import *

###########################################

NPROCESS = 4

ribo_output_file = "cli_create.ribo"


class TestCLIBase(unittest.TestCase):

    def run_command(self, command):
      """
      Command must be a list.
      See examples below,
      """

      command_str    = tuple( map( str, command ) )
      process        = subprocess.Popen(command_str, 
                                        stdout = subprocess.PIPE,
                                        stderr = subprocess.PIPE)
      output, error  = process.communicate()
      output_str     = output.decode()
      error_str      = error.decode()

      return (output_str, error_str) 

    def setUp(self):
        self.tmp_files = list()

        # We needed this adjustment for bedToBam compatiblitiy
        # Our bed file has whitespace instead of tabs
        # thios makes bedToBam complain
        self.TRANSCRIPT_LENGTHS_str = TRANSCRIPT_LENGTHS.replace(" ","\t")
        self.READ_SET_1_str         = READ_SET_1.replace("   ", "\t")
        self.READ_SET_1_str         = self.READ_SET_1_str.replace("  ", "\t")
        self.READ_SET_1_str         = self.READ_SET_1_str.replace(" ", "\t")


        files_and_contents = \
            [ (REF_LEN_FILE,     self.TRANSCRIPT_LENGTHS_str),
              (ANNOTATION_FILE,  TRANSCRIPT_ANNOTATION),
              (ALIGNMENT_FILE_1, self.READ_SET_1_str),
              (METADATA_FILE_1,  METADATA_EXPERIMENT_STR_1),
              (RIBO_META_FILE_1, RIBO_METADATA_STR_1),
              (RNASEQ_FILE_1,    RNASEQ_DATA_1) ]

        for t_file, content_str in files_and_contents:
            with open(t_file, "w") as output_stream:
                print(content_str, file = output_stream)
            self.tmp_files.append(t_file)

        if os.path.isfile(ribo_output_file):
            os.remove(ribo_output_file)

        create_command = ["ribog", "create", 
                          "--name",           "merzifon",
                          "--alignmentfile",  ALIGNMENT_FILE_1, 
                          "--reference",      "hg38",
                          "--lengths",        REF_LEN_FILE, 
                          "--annotation",     ANNOTATION_FILE,
                          "--metageneradius", METAGENE_RADIUS, 
                          "--leftspan",       LEFT_SPAN, 
                          "--rightspan",      RIGHT_SPAN,
                          "--lengthmin",      2, 
                          "--lengthmax",      5, 
                          "--expmeta",        METADATA_FILE_1,
                          "--ribometa",       RIBO_META_FILE_1,
                          "--rnaseq",         RNASEQ_FILE_1,
                          "-n",               4,
                          ribo_output_file]

        output, error      = self.run_command(create_command)

        #print("Output:", output)
        self.tmp_files.append( ribo_output_file )
        if output:
            print(output)
        if error:
            print(error)


    def tearDown(self):
        pass
        [ os.remove(f) for f in self.tmp_files ]

#######################################################################        

#@unittest.skip("temporarily skipping plot tests")
class TestCLICreate(TestCLIBase):

    ######################################################
    def test_create_ribo_exists(self):
        self.assertTrue(os.path.exists(ribo_output_file))
        myribo             = h5py.File(ribo_output_file, "r")
        experiments_handle = myribo[EXPERIMENTS_name]
        experiment_names   = tuple(experiments_handle.keys())
        myribo.close()

        self.assertTrue("merzifon" in experiment_names)



    def test_info_display(self):
        command_pieces = [ "ribog", "info" , ribo_output_file ]
        output, error  = self.run_command(command_pieces)
        if error:
            print("error: ", error)

        self.assertTrue( "merzifon" in output )
        self.assertTrue( "Version" in output )


    def test_rnaseq(self):
        command_pieces = [ "ribog", "rnaseq" , "get", 
                           "--name", "merzifon", ribo_output_file ]
        output, error  = self.run_command(command_pieces)

        output_lines = output.split("\n")

        t_name, t_exp = output_lines[2].split()

        self.assertTrue(np.isclose( float(t_exp), 15.46 ))
        self.assertEqual(t_name, "VEGFA")


    def test_exp_name_from_path(self):
        """
        If the name is not provided,
        it can be extracted from the filepath. 
        """

        ribo_experiment_file = "this_file.ribo"
        self.tmp_files.append(ribo_experiment_file)

        create_command = ["ribog", "create", 
                          "--alignmentfile",  ALIGNMENT_FILE_1, 
                          "--reference",      "hg38",
                          "--lengths",        REF_LEN_FILE, 
                          "--annotation",     ANNOTATION_FILE,
                          "--metageneradius", METAGENE_RADIUS, 
                          "--leftspan",       LEFT_SPAN, 
                          "--rightspan",      RIGHT_SPAN,
                          "--lengthmin",      2, 
                          "--lengthmax",      5, 
                          "--expmeta",        METADATA_FILE_1,
                          "--ribometa",       RIBO_META_FILE_1,
                          "--rnaseq",         RNASEQ_FILE_1,
                          "-n",               4,
                          ribo_experiment_file]

        output, error  = self.run_command(create_command)

        with h5py.File(ribo_experiment_file , "r") as ribo:
            experiment_list = tuple( ribo[EXPERIMENTS_name].keys() )
            self.assertTrue( "this_file" in experiment_list )


    def test_create_from_bam_file(self):
        """
        Test if we can create ribo files directly from bam files
        """
        if not shutil.which('bamToBed'):
            raise FileNotFoundError("Could not find the executable bamToBed")
        if not shutil.which('bedToBam'):
            raise FileNotFoundError("Could not find the executable bedToBam")

        ribo_experiment_file = "from_bam_file.ribo"
        self.tmp_files.append(ribo_experiment_file)

        ALIGNMENT_FILE_BAM_1 = "frombed.bam"
        self.tmp_files.append(ALIGNMENT_FILE_BAM_1)

        # First we make the bam file
        pipe_t = pipes.Template()
        pipe_t.append("bedToBam -g {}"\
                      .format( REF_LEN_FILE), "--")
      
        f = pipe_t.open(ALIGNMENT_FILE_BAM_1, 'w')
        f.write(self.READ_SET_1_str)
        f.close()
        
        if not os.path.isfile(ALIGNMENT_FILE_BAM_1):
            raise FileNotFoundError("Couldnt convert bed file to bam.")

        create_command = ["ribog", "create", 
                          "--name",            "merzifon",
                          "--alignmentfile",   ALIGNMENT_FILE_BAM_1,
                          "--alignmentformat", "bam", 
                          "--reference",       "hg38",
                          "--lengths",         REF_LEN_FILE, 
                          "--annotation",      ANNOTATION_FILE,
                          "--metageneradius",  METAGENE_RADIUS, 
                          "--leftspan",        LEFT_SPAN, 
                          "--rightspan",       RIGHT_SPAN,
                          "--lengthmin",       2, 
                          "--lengthmax",       5, 
                          "--expmeta",         METADATA_FILE_1,
                          "--ribometa",        RIBO_META_FILE_1,
                          "--rnaseq",          RNASEQ_FILE_1,
                          "-n",                4,
                          ribo_experiment_file]

        output, error  = self.run_command(create_command)

        with h5py.File(ribo_experiment_file , "r") as ribo:
            experiment_list = tuple( ribo[EXPERIMENTS_name].keys() )
            self.assertTrue( "merzifon" in experiment_list )



class TestCLIMetadata(TestCLIBase):

    
    #######################################################


    def test_ribo_metadata(self):
        command_pieces = [ "ribog", "metadata" , "get", 
                            ribo_output_file ]
        output, error  = self.run_command(command_pieces)

        self.assertTrue("bowtie2" in output)


    def test_delete_ribo_metadata(self):
        command_pieces = [ "ribog", "metadata" , "delete", 
                            "--force", ribo_output_file ]
        output, error  = self.run_command(command_pieces)

        command_pieces = [ "ribog", "metadata" , "get", 
                            ribo_output_file ]
        output, error  = self.run_command(command_pieces)

        self.assertTrue("bowtie2" not in output)


    def test_experiment_metadata(self):
        command_pieces = [ "ribog", "metadata" , "get", 
                           "--name", "merzifon",
                           ribo_output_file ]
        output, error  = self.run_command(command_pieces)

        self.assertTrue("HeLa" in output)


    def test_delete_experiment_metadata(self):
        command_pieces = [ "ribog", "metadata" , "delete", 
                           "--name", "merzifon",
                            "--force", ribo_output_file ]
        output, error  = self.run_command(command_pieces)

        command_pieces = [ "ribog", "metadata" , "get", 
                           "--name", "merzifon",
                            ribo_output_file ]
        output, error  = self.run_command(command_pieces)

        self.assertTrue("HeLa" not in output)
        self.assertTrue(not output)
        self.assertTrue(not error)


    def test_change_ribo_metadata(self):

        new_ribo_meta_file = "new_ribo_meta_file.yaml"
        with open(new_ribo_meta_file, "w") as meta_out_stream:
            print("new_aligner: bwa", file = meta_out_stream)

        self.tmp_files.append(new_ribo_meta_file)

        command_pieces = [ "ribog", "metadata" , "set",
                           "--meta",   new_ribo_meta_file, 
                            "--force", ribo_output_file ]
        output, error  = self.run_command(command_pieces)

        command_pieces = [ "ribog", "metadata" , "get", 
                            ribo_output_file ]
        output, error  = self.run_command(command_pieces)

        self.assertTrue("bwa" in output)

    def test_change_experiment_metadata(self):

        new_exp_meta_file = "new_exp_meta_file.yaml"
        with open(new_exp_meta_file, "w") as meta_out_stream:
            print("new_type: HEK293", file = meta_out_stream)

        self.tmp_files.append(new_exp_meta_file)

        command_pieces = [ "ribog", "metadata" , "set",
                           "--meta", new_exp_meta_file, 
                           "--name", "merzifon",
                            "--force", ribo_output_file ]
        output, error  = self.run_command(command_pieces)

        command_pieces = [ "ribog",  "metadata" , "get", 
                           "--name", "merzifon",
                            ribo_output_file ]
        output, error  = self.run_command(command_pieces)

        self.assertTrue("HEK293" in output)



###########################################################

if __name__ == '__main__':
        
    unittest.main()


# -*- coding: utf-8 -*-
from collections import OrderedDict

import h5py
import numpy as np

from ..settings import *

###############################################################

REF_INDEX        = 0
FIVE_P_END_INDEX = 1


def find_coverage( input_bed_stream,  
                   ref_names , ref_lengths):
    """
    Computes the coverage per transcript.

    Note that the order in the ref_names and the ref_lengths must be compatible.

    This order is also kept in the order of the keys of the
    returned value ( coverage_dict )

    Parameters
    ----------
    input_bed_stream: IO_stream
       Alignemt reads are in bed format.
       This is a handle to the bed file

    ref_names: list(str)
       List of transcript names

    ref_lengths: list(int)
       List of transcript lengths.
       They are in the same order with the ref names.

    Returns
    -------
    coverage_dict: OrderedDict(numpy.array)
       The keys are transcript names in the same
       order as ref_names
       The values are numpy arrays corresponding
       to coverage at the given index position.

    """
    assert len(ref_names) == len(ref_lengths)

    coverage_dict = OrderedDict()

    for ref_name, ref_length in zip(ref_names, ref_lengths):
        coverage_dict[ref_name] = np.zeros( ref_length, 
                                            dtype = TRANSCRIPT_COVERAGE_DT) 

    for entry in input_bed_stream:
        contents = entry.strip().split()
        if len(contents) < 6:
            continue
        this_ref    = contents[REF_INDEX]
        this_five_p = int(contents[FIVE_P_END_INDEX])
        coverage_dict[this_ref][ this_five_p ] += 1

    return coverage_dict


# -*- coding: utf-8 -*-

import numpy as np
from ..settings import *

def get_extended_boundaries(annotation, left_span, right_span):
    extended_annotation = list()

    """
        Note that a is of the form
         ( (UTR5_orig_start, UTR5_orig_end),
           (CDS_orig_start, CDS_orig_end),
           (UTR3_orig_start, UTR3_orig_end)
          )
    """
    for a in annotation:

        UTR5_start = 0
        UTR5_junction_start = max(a[1][0] - left_span, 0)
        UTR5_end = UTR5_junction_start
        UTR5_junction_end   = min(a[1][0] + right_span + 1 , a[1][1] )
        CDS_start = UTR5_junction_end
        CDS_end = max( a[1][1] - left_span, CDS_start )
        UTR3_junction_start = CDS_end
        UTR3_junction_end = min( a[1][1] + right_span + 1 , a[2][1])
        UTR3_start = UTR3_junction_end
        UTR3_end   = a[2][1] 

        extended_annotation.append( ( (UTR5_start, UTR5_end),
                                      (UTR5_junction_start, UTR5_junction_end),
                                      (CDS_start, CDS_end),
                                      (UTR3_junction_start, UTR3_junction_end),
                                      (UTR3_start, UTR3_end) ) 
                                  )

    return extended_annotation

def find_region_counts( coverage, annotation, left_span, right_span ):
    """
    Finds the number of counts for each region.

    Each transcript is partitioned into 5 regions:
    UTR5, UTR5_junction, CDS, UTR3_junction and UTR3

    The boundaries of these regions are as follows:
    Note that the second coordinate component is excluded

    UTR5 = [0, start_site - left_span ) 
    UTR5_junction: [start_site - left_span, start_site + right_span + 1 )
    CDS = [ start_site + right_span + 1, stop_site - left_span )
    UTR3_junction: [stop_site - left_span, stop_site + right_span + 1 )
    UTR3: [ stop_site + right_span  +1, REF_length)

    Parameters
    ----------
    coverage: This is an ordered dict where 
              each keyword is a reference name each value is an array
              whose length is the nucleotide length of the refrence
              This array has the coverage of the given reference 
              in nucleotide level.
              So, for example, coverage["MYC"][10] 
              gives the number of nucleotides mapped to the 10th position
              of the MYC gene.

    left_span: Number of nucleotides to the left of the start / stop site.
               This is used to determine the junction regions.
            
    annotation: An list of triplets where each triplet is of the form
                (  [UTR3_star, UTR3_END], 
                   [CDS_start, CDS_end], 
                   [UTR5_Start, UTR5_end]) 

    Returns
    -------
    region_counts: a 2-D numpy array of dimension #references X 5
                   each coulmn corresponds to a region
    """

    extended_region_boundaries = \
            get_extended_boundaries(annotation, left_span, right_span)

    # initialize all region counts to 0
    # Note that 5 is for the 5 regions of interest
    # UTR5, UTR5_junction, CDS, UTR3_junction, UTR3
    region_counts = np.zeros(shape = (len(annotation), 5), dtype = np.uint32)

    # Since the two lists are ordered the same way
    # we can safely zip them
    # Note that coverage is coming from and ordereddict
    iter_items = zip( extended_region_boundaries, 
                      coverage.values(), 
                      list(range(len(annotation))) )

    for boundaries, cover, ref_index in iter_items:
        for region_ind, start_stop in enumerate(boundaries):
            region_counts[ref_index, region_ind] = \
                np.add.reduce( cover[ start_stop[0]: start_stop[1] ] )

    return region_counts




# ARTIFICIAL test data for testing ribogadgets


REF_LEN_FILE = "ref_lengths.tsv"
TEST_RIBO = "test_create.ribo"
ANNOTATION_FILE = "annotation.bed"
ALIGNMENT_FILE_1 = "alignment_file_1.bed"

###########################################


TRANSCRIPT_LENGTHS=\
"""GAPDH	1290
VEGFA	875
MYC	1462
BRCA	565"""


TRANSCRIPT_ANNOTATION=\
"""GAPDH    0   50  UTR5    0   -
GAPDH   50  1130    CDS 0   -
GAPDH   1130    1290 UTR3   0   -
VEGFA   0   675 CDS 0   +
VEGFA   675 875 UTR3    0   +
MYC 0   90  UTR5    0   -
MYC 90  1251    CDS 0   -
MYC 1251    1462    UTR3    0   -
BRCA    0   2   UTR5    0   +
BRCA    2   561 CDS 0   +
BRCA    561 565 UTR3    0   +"""

READ_SET_1=\
"""GAPDH 250  276  read_1 0   +
BRCA 1  33  read_2 0 +"""

GENERAL_METADATA=\
{
    "metagene_radius": 5,
    "left_span": 3,
    "right_span": 2
}

RNASEQ_DATA_1 = \
"GAPDH\t2.89\nVEGFA\t15.46\nMYC\t8"


# -*- coding: utf-8 -*-
import unittest

import os
from io import StringIO, BytesIO

import numpy as np
import pandas as pd
import h5py


from ribogadgets import create
from ribogadgets.core.coverage import find_coverage
from ribogadgets.core.get_gadgets import get_reference_names,\
                                         get_reference_lengths,\
                                         get_region_boundaries
from ribogadgets.settings import *
from ribogadgets.dump import *
from ribogadgets.rnaseq import *

from ribogadgets.merge import merge_ribos

from multilength_test_data import *
from ribogadgets.core import create_experiment

###########################################

NPROCESS = 4

class TestCreate(unittest.TestCase):

    def setUp(self):
        self.ref_len_file    = StringIO(TRANSCRIPT_LENGTHS)
        self.annotation_file = StringIO(TRANSCRIPT_ANNOTATION)
        self.alignment_file  = StringIO(READ_SET_1)

        self.handle = h5py.File(BytesIO())

        create.create_ribo(     self.handle, "merzifon", 
                                alignment_file  = self.alignment_file,
                                reference_name  = "appris_human_v2",
                                length_min      = 2,
                                length_max      = 5,
                                metagene_radius = METAGENE_RADIUS,
                                left_span       = LEFT_SPAN, 
                                right_span      = RIGHT_SPAN,
                                lengths_file    = self.ref_len_file,
                                store_coverage  = True,
                                annotation_file = self.annotation_file)

    def tearDown(self):
        self.handle.close()

    def initial_rnaseq(self):
        rnaseq_lines = RNASEQ_DATA_1.split("\n")
        rnaseq_expression  = tuple( map(lambda x: x.split()[1] , rnaseq_lines) )
        rnaseq_expression  = np.array(rnaseq_expression, dtype = RNASEQ_DT)
        rnaseq_transcripts = tuple( map(lambda x: x.split()[0] , rnaseq_lines) )
        rnaseq_df = pd.DataFrame( rnaseq_expression, index = rnaseq_transcripts )

        set_rnaseq(self.handle, "merzifon", rnaseq_df)

    ### T E S T   R N A S E Q  ################################

    def test_set_rnaseq(self):
        self.initial_rnaseq()

        stored_rnaseq =  self.handle[EXPERIMENTS_name]\
                            ["merzifon"][RNASEQ_name][RNASEQ_name][...]
        self.assertTrue( all(np.isclose( [2.89, 15.46, 8], stored_rnaseq ) ) )

        self.assertTrue(not all(np.isclose( [2.89, 15.46, 8.25], stored_rnaseq ) ) )

        expression_2    = np.array([0.5, 983.972] , dtype = RNASEQ_DT)
        t_list_2        = ["MYC", "GAPDH"]
        rnaseq_df_2     = pd.DataFrame( expression_2, index = t_list_2 )

        set_rnaseq(self.handle, "merzifon", rnaseq_df_2)
        stored_rnaseq_2 =  self.handle[EXPERIMENTS_name]\
                            ["merzifon"][RNASEQ_name][RNASEQ_name][...]

        # Note that the order changes and VEGFA has 0
        self.assertTrue( all(np.isclose( [983.972, 0, 0.5], stored_rnaseq_2 ) ) )


    def test_get_rnaseq(self):
        self.initial_rnaseq()

        read_rnaseq = get_rnaseq(self.handle, "merzifon")

        read_transcripts = tuple(read_rnaseq.index)
        read_expression  = read_rnaseq["merzifon"]

        self.assertEqual( read_transcripts[0], "GAPDH" )
        self.assertEqual( read_transcripts[2], "MYC" )

        self.assertTrue( all(np.isclose( [2.89, 15.46, 8], read_expression ) ) )


    def test_delete_rnaseq(self):
        self.initial_rnaseq()

        self.assertTrue(RNASEQ_name in \
                         tuple(self.handle[EXPERIMENTS_name]["merzifon"].keys() ))

        delete_rnaseq(self.handle, "merzifon")

        self.assertTrue(RNASEQ_name not in \
                         tuple(self.handle[EXPERIMENTS_name]["merzifon"].keys() ))






class TestRNASEQFROMMERGED(unittest.TestCase):

    def setUp(self):
        self.tmp_files = list()


        self.ref_len_file       = StringIO(TRANSCRIPT_LENGTHS)
        self.annotation_file    = StringIO(TRANSCRIPT_ANNOTATION)
        self.alignment_file_1   = StringIO(READ_SET_1)
        self.alignment_file_2   = StringIO(READ_SET_2)

        self.handle   = h5py.File(BytesIO() )
        self.handle_2 = h5py.File(BytesIO() )

        create.create_ribo(
                ribo            = self.handle, 
                experiment_name = "merzifon", 
                alignment_file  = self.alignment_file_1,
                reference_name  = "hg38",
                lengths_file    = self.ref_len_file, 
                annotation_file = self.annotation_file,
                metagene_radius = METAGENE_RADIUS, 
                left_span       = LEFT_SPAN, 
                right_span      = RIGHT_SPAN,
                length_min      = 2,
                length_max      = 5,  
                nprocess        = NPROCESS,
                tmp_file_prefix = "")
        self.ref_len_file.seek(0)
        self.annotation_file.seek(0)
        self.alignment_file_1 = StringIO(READ_SET_1)

        
        create.create_ribo(
                ribo            = self.handle_2, 
                experiment_name = "ankara", 
                alignment_file  = self.alignment_file_2,
                reference_name  = "hg38",
                lengths_file    = self.ref_len_file, 
                annotation_file = self.annotation_file,
                metagene_radius = METAGENE_RADIUS, 
                left_span       = LEFT_SPAN, 
                right_span      = RIGHT_SPAN,
                length_min      = 2,
                length_max      = 5,
                nprocess        = NPROCESS,
                tmp_file_prefix = "")
        self.ref_len_file.seek(0)
        self.annotation_file.seek(0)
        self.alignment_file_2 = StringIO(READ_SET_2)

        self.merged_ribo_without_rnaseq = h5py.File(BytesIO(), "w")
        merge_ribos( self.merged_ribo_without_rnaseq, 
                     [self.handle , self.handle_2] )


        rnaseq_lines = RNASEQ_DATA_1.split("\n")
        rnaseq_expression  = tuple( map(lambda x: x.split()[1] , rnaseq_lines) )
        rnaseq_expression  = np.array(rnaseq_expression, dtype = RNASEQ_DT)
        rnaseq_transcripts = tuple( map(lambda x: x.split()[0] , rnaseq_lines) )
        rnaseq_df = pd.DataFrame( rnaseq_expression, index = rnaseq_transcripts )

        set_rnaseq(self.handle, 
                   name      = "merzifon", 
                   rnaseq_df = rnaseq_df)

        rnaseq_lines = RNASEQ_DATA_2.split("\n")
        rnaseq_expression  = tuple( map(lambda x: x.split()[1] , rnaseq_lines) )
        rnaseq_expression  = np.array(rnaseq_expression, dtype = RNASEQ_DT)
        rnaseq_transcripts = tuple( map(lambda x: x.split()[0] , rnaseq_lines) )
        rnaseq_df = pd.DataFrame( rnaseq_expression, index = rnaseq_transcripts )

        set_rnaseq(self.handle_2, 
                   name      = "ankara", 
                   rnaseq_df = rnaseq_df)

        self.merged_ribo = h5py.File(BytesIO(), "w")
        merge_ribos( self.merged_ribo, [self.handle , self.handle_2] )

    def test_from_merged(self):

        all_rnaseq = get_rnaseq(self.merged_ribo)

        self.assertTrue( all(np.isclose(all_rnaseq["merzifon"] , [2.89, 15.46, 8]) ) )
        self.assertTrue( all(np.isclose(all_rnaseq["ankara"] , [6.42, 1.37, 0.06]) ) )

    def test_from_no_rnaseq(self):

        all_no_rnaseq = get_rnaseq(self.merged_ribo_without_rnaseq)

        self.assertTrue( not all_no_rnaseq)


if __name__ == '__main__':
        
    unittest.main()


from .main import *
import click
from ..info import ribo_file_info
                   
###########################################################

@cli.command()
@click.argument('ribofile', type=click.Path(exists=True ))
def info(ribofile):
    """
    Displays a summary information about the given ribo file
    """

    display_str = ribo_file_info(ribo_file = ribofile)

    print( display_str)





# -*- coding: utf-8 -*-

import numpy as np
from ..settings import *


def find_site_coverage( coverage, radius, annotation, site_type ):
    """
    Finds the coverage of a given site ( start / stop) for a given radius
    In ribosome-profiling jargon, this is  known as "metagene analysis".

    Parameters
    ----------
    coverage: This is an ordered dict where 
              each keyword is a reference name each value is an array
              whose length is the nucleotide length of the refrence
              This array has the coverage of the given reference 
              in nucleotide level.
              So, for example, coverage["MYC"][10] 
              gives the number of nucleotides mapped to the 10th position
              of the MYC gene.

    radius: Number of nucleotides to the left and right of the site of interest.
            
    annotation: An list of triplets where each triplet is of the form
                (  [UTR3_start, UTR3_END], [CDS_start, CDS_end], [UTR5_Start, UTR5_end]) 

    site_type: Valid choices are "start" and "stop"
               Finds the coverage around the start or stop site accordingly

    Returns
    -------
    2-dimensional numpy array of size (#references) X ( (2*radius) + 1 )
    Each row corresponds to one reference and each column corresponds to a
    nucleotide position (relative to the site). The site position will be at the middle column.
    

    Notes
    -----
    Note that by our design, the order of references in the 
    coverage OrderedDict and in annotation must be the same.
   
    """   

    if site_type == "start":
        ref_coordinates = list( map(lambda x: (x[0][0], x[1][0], x[2][1]) , annotation ) )
    elif site_type == "stop":
        ref_coordinates = list( map(lambda x: (x[0][0], x[1][1], x[2][1]) , annotation ) )
    else:
        raise IOError("Site type can be either start or stop."
                      "Provided type is " + site_type)

    array_dimensions = (len(ref_coordinates), 2*radius + 1)
    site_coverage = np.zeros(shape = array_dimensions, dtype = SITE_COVERAGE_DT)

    for ref_index, ref_items in enumerate(list(coverage.items() )) :
        ref_name, ref_coverage = ref_items
        this_lower_coord, this_site_coord, this_upper_coord = \
                   ref_coordinates[ref_index]

        index_range_lower = this_site_coord - radius
        index_range_upper = this_site_coord + radius + 1
        index_range_raw = np.arange(index_range_lower, index_range_upper)

        index_range_logical = np.all( [index_range_raw >= this_lower_coord ,\
                               index_range_raw < this_upper_coord ] ,\
                               axis = 0)
         
        this_coverage = site_coverage[ref_index, :]
        for k in range( 2*radius + 1 ):
            if index_range_logical[k]:
                this_coverage[k] = coverage[ref_name][this_site_coord  - radius + k]

        site_coverage[ref_index, :] = this_coverage
  

    return site_coverage



# -*- coding: utf-8 -*-

import functools

import h5py 

from .get_gadgets import *

from .exceptions import ExperimentDoesntExist


#####################################################################
## TODO ?? Revise this function
def prompt_user(message, attr_exists, force):
    if not attr_exists or force:
        return True
    else:
        print(message)
        user_response = input("Do you want to continue? y/n  : ")

        if user_response.lower().startswith("y"):
            return True
        else:
            exit(1)
    


def handle_error(error_message, explanation_message):
    if not error_message:
        return 0
    else:
        error_message += "\n" + explanation_message
        raise ValueError(error_message)

"""
def verify_read_length_range( ribo_handle, 
                              range_lower, range_upper ):
    
    error_message = ""
    length_min, length_max = get_read_length_range(ribo_handle)
    explanation_message    = " ".join(["This ribo file has min read length:",
                                     str(length_min), "and max read length:", 
                                     str(length_max), "."  ])

    if range_lower < length_min or range_lower > length_max:
        error_message = \
          "Given lower range {} is not in the correct interval.".\
             format(range_lower) 

    if range_upper > length_max or range_upper < length_min:
        error_message = \
          "Given upper range {} is not in the correct interval.".\
             format(range_upper)

    return handle_error(error_message, explanation_message) 
"""

"""

def verify_experiment_names (ribo_handle, experiment_names):

    existing_experiment_names = get_experiment_names(ribo_handle)
    error_message = ""
    explanation_message = "Existing experiments are: \n"+\
                          ", ".join(existing_experiment_names)

    for s in experiment_names:
        if s not in existing_experiment_names:
            error_message += \
              "The experiment {} does not exist.\n".format(s)

    return handle_error(error_message, explanation_message)
"""

def verify_extended_region_name(region_name):
    if region_name not in EXTENDED_REGION_names:
        error_message = "No region exists with the name " + region_name
        error_message += " Available regions are:" + \
                          join(", ".EXTENDED_REGION_names)
        raise ValueError(error_message)

def verify_site_type(site_type):
    valid_site_types = (REF_DG_START_SITE_COV, REF_DG_STOP_SITE_COV)
    if site_type not in valid_site_types:
        error_message = "You provided {} as site type.\n".format(site_type)
        error_message+= "Site type can be either {} or {}.".format(
                          *valid_site_types)
        raise ValueError(error_message)


###########################################################
##  D E C O R A T O R S
###########################################################


def make_cli_function(func):
    """
    This function decorator catches ribogadgets specific errors and
    prints them on the standard error
    """
    @functools.wraps(func)
    def cli_func_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except RiboBaseError as e:
            exit(e)
    return cli_func_wrapper


def check_experiment_list_in_ribo_handle(func):
    """
    NOTE THAT THIS DECORATOR IMPOSES CERTAIN
    NAMING CONVENTIONS ON THE ARGUMENTS OF func.

    If there are any experiments that does not exist in the
    ribo, raises an exception.

    Assumptions:
    The first argument of the decorated fucntion is ribo handle
    or it has ribo_handle as a keyword argument

    Experiemnts are given as a list in "experiemnt_list"
    or one experiment as 
    """
    @functools.wraps(func)
    def check_experiments_wrapper(*args, **kwargs):

        # Ribo file is either the first argument
        # or given in the keyword argument entitled "ribo_file"
        ribo_handle = kwargs.get("ribo_handle", None)
        if not ribo_handle:
            ribo_handle = args[0]

        # Either there is an experiment_list
        # or experiment_name
        # or just name in the keyword arguments
        # so we construct experiment list accordingly
        experiment_list       = []
        ribo_experiments      = []
        not_found_experiments = []

        if kwargs.get("experiment_list", None):
            experiment_list = kwargs["experiment_list"]
        elif kwargs.get("experiment_name", None):
            experiment_list = [ kwargs["experiment_name"] ]
        elif kwargs.get("name", None):
            experiment_list = [ kwargs["name"] ]

        if len(experiment_list) > 0:
            ribo_experiments = get_experiment_names(ribo_handle)

        for e in experiment_list:
            if e not in ribo_experiments:
                not_found_experiments.append(e)

        if len(not_found_experiments) > 0:
            plural_1, plural_2 = "", ""

            if len(not_found_experiments) > 1 :
                plural_1 = "s"
                plural_2 = "es" 

            exp_str = ", ".join(not_found_experiments)
            error_message = "The following experiment{} do{} not exist:\n"\
                               .format(plural_1, plural_2) + \
                               exp_str
            raise ExperimentDoesntExist(error_message)

        return func(*args, **kwargs)

    return check_experiments_wrapper


def check_length_range(func):
    """
    Compares the boundaries of the arguments
    to the minimum and maximum length of the ribo file

    If these ranges are not "compatible",
    raises an Exception.

    Assumption:
    -----------
    kwargs of the function decorated must have the keys
    range_lower and range_upper
    """
    @functools.wraps(func)
    def check_length_range_wrapper(*args, **kwargs):
        ribo_handle = kwargs.get("ribo_handle", None)
        if not ribo_handle:
            ribo_handle = args[0]  

        min_read_length, max_read_length = \
            get_read_length_range(ribo_handle)


        # If either of the boundaries are not set,
        # set them to ribo length boundaries.
        if not kwargs["range_lower"] :
            kwargs["range_lower"] = min_read_length
        if not kwargs["range_upper"] :
            kwargs["range_upper"] = max_read_length


        range_lower = kwargs["range_lower"]
        range_upper = kwargs["range_upper"]

        error_message = ""

        # We don't expect this but we check it just in case.
        if max_read_length < min_read_length:
            error_message = "There is a problem with the read length" + \
                            " range of this ribo file. Minimum read length " + \
                            "is greater than maximum length."
            error_message += "Min Read Length = {},\n Max Read Length = "\
                .format(min_read_length, max_read_length)

        if range_lower > range_upper:
            error_message = "Range Lower = {} can not be larger than " + \
                            "Range Upper = {}".format(range_upper, range_upper)

        if range_lower < min_read_length:
            error_message = "Minimum Length can not be less than {}"\
                              .format(min_read_length)

        if range_upper > max_read_length:
            error_message = "Maximumu read length can not be more than {}"\
                              .format(max_read_length)

        if error_message:
            raise InvalidLengthRange(error_message)
        else:
            return func(*args, **kwargs)

    return check_length_range_wrapper




# -*- coding: utf-8 -*-

import click

from .main import *

from ..dump import *

##################################

@cli.group()
def dump():
    """Dump selected parts of ribo files to particular formats"""
    pass


@dump.command()
@click.argument("ribo", type = click.Path(exists = True))
@click.option('-s', '--site',
              type      = click.Choice(['start', 'stop']),
              required  = True,
              help      = 'Site type.')
@click.option('-o', '--out',
              help      = 'Output file in bed format')
@click.option('-e', '--experiment',
              help      = 'Name of the experiment')
@click.option( '-l', '--lowerlength',
               type     = click.INT,
               help     = 'Minimum read length to be taken')
@click.option( '-u', '--upperlength',
               type     = click.INT,
               help     = 'Maximum read length to be taken')
@click.option( '--sumlengths',
                is_flag = True,
                help    = 'Sum accross lengths')
@click.option( '--nosumtrans',
               is_flag  = True,
               help     = "Do NOT aggregate values"
                          " accross transcripts")
def metagene(ribo, 
             site, 
             out, 
             experiment, 
             lowerlength, upperlength, 
             sumlengths, 
             nosumtrans ):
    """Dumps metagene data 
    to a csv file or standard output.

    Metagene data is obtained from
    coverage around start / stop site in a pre-defined radius.


    Examples:
     Below are some examples for different scenarios.
     
     1) Get metagene data around START site for the experiment WT.
        Report results for read lengths from 28 to 32. 
        Agrregate data by summing accros read lengths.
        Also, values are summed across transcripts.
        Save the results in start.csv.
          
            \b 
            
            .. code:: bash      
             
             ribog dump metagene \\
                --experiment WT \\
                --site start \\
                --out start.csv \\
                --lowerlength 28 --upperlength 32 \\
                --sumlengths \\
                sample.ribo
                
     2) Get metagene data around STOP site for the experiment Treatment_1.
        Report results for read length 31. 
        Values are summed across transcripts.
        Print the results on the standard output.
          
            \b 
            
            .. code:: bash      
             
             ribog dump metagene \\
                --experiment WT \\
                --site stop \\
                --lowerlength 31 --upperlength 31 \\
                sample.ribo
                
     3) Get metagene data around START site for the experiment WT.
        Report results for read lengths from 30 to 32. 
        Report results for each read length.
        Also, values are summed across transcripts.
        Save the results in start.csv.
          
            \b 
            
            .. code:: bash      
             
             ribog dump metagene \\
                --experiment WT \\
                --site start \\
                --out start.csv \\
                --lowerlength 30 --upperlength 32 \\
                sample.ribo
                
    """

    experiment_list = []
    if experiment:
        experiment_list = [experiment]

    return dump_metagene_wrapper( 
                           ribo_file       = ribo, 
                           output_file     = out,
                           site_type       = site, 
                           sum_lengths     = sumlengths, 
                           sum_references  = (not nosumtrans),
                           range_lower     = lowerlength, 
                           range_upper     = upperlength,
                           experiment_list = experiment_list)




###########################################################

@dump.command()
@click.argument("ribo",   type = click.Path(exists=True))
@click.option('-r', '--region',
              type      = click.Choice(['UTR5', 'UTR5_junction',
                                        'CDS',
                                        'UTR3', 'UTR3_junction']),
              required  = True,
              help      = "Site type.")
@click.option('-o', '--out',
              help      = 'Output file in csv format')
@click.option('-e', '--experiment',
              help      = 'Name of the experiment')
@click.option( '--lowerlength',
               type     = click.INT,
               help     = 'Minimum read length to be taken')
@click.option( '--upperlength',
               type     = click.INT,
               help     = 'Maximum read length to be taken')
@click.option( '--sumlengths',
                is_flag = True,
                help    = 'Sum accross lengths')
@click.option( '--sumtrans',
               is_flag  = True,
               help     = 'Sum accross transcripts')
def region_counts(ribo, 
                  region, 
                  out, 
                  experiment,
                  lowerlength, 
                  upperlength, 
                  sumlengths, 
                  sumtrans ):
    """
    Dumps a given region in csv format to a file or standard output.
    
    Examples:
     Below are some examples for different scenarios.
     
     1) Get number of reads mapping to the coding sequence (CDS) 
        for the experiment WT.
        Report results for read lengths from 28 to 32. 
        Agrregate data by summing accros read lengths.
        Also, values are summed across transcripts.
        Save the results in cds.csv.
          
            \b 
            
            .. code:: bash      
             
             ribog dump region-counts \\
                --experiment WT \\
                --region CDS \\
                --out cds.csv \\
                --lowerlength 28 --upperlength 32 \\
                --sumlengths \\
                --sumtrans \\
                sample.ribo
                
     2) Get number of reads mapping to UTR3 
        for the experiment Treatment.
        Report results for read lengths from 30 to 32. 
        Agrregate data by summing accros read lengths.
        CDS occupancy is reported for each transcript.
        Save the results in cds.csv.
          
            \b 
            
            .. code:: bash      
             
             ribog dump region-counts \\
                --experiment Treatment \\
                --region UTR3 \\
                --out cds.csv \\
                --lowerlength 30 --upperlength 32 \\
                --sumlengths \\
                sample.ribo
    """

    experiment_list = []
    if experiment:
        experiment_list = [experiment]

    return dump_region_wrapper(
                        ribo_file       = ribo, 
                        output_file     = out, 
                        region_name     = region,
                        sum_lengths     = sumlengths, 
                        sum_references  = sumtrans, 
                        range_lower     = lowerlength, 
                        range_upper     = upperlength,
                        experiment_list = experiment_list)


######################################################

@dump.command()
@click.argument("ribo", type = click.Path(exists = True))
@click.option('-o', '--out',
              help = 'Output file in bed format')
def annotation(ribo, out):
    """Gives UTR5, CDS and UTR3 annotation in bed format.

    Annotation is written to standard output if no
    output file is given.

    Examples:
     1) Store the annotation in a bed file 
     

     .. code:: bash
         
         ribog dump annotation -o regions.bed sample.ribo
         
     2) Print annotation to standard output 
     
     .. code:: bash
         
        ribog dump annotation sample.ribo
         
     
    """
    return dump_annotation_wrapper(ribo, out)


@dump.command()
@click.argument("ribo", type = click.Path(exists=True))
@click.option('-o', '--out',
              help     = 'Output file',
              type     = click.Path(False),
              required = False)
@click.option('-s', '--sep',
              help            = 'Column separator, default is \\t (tab)',
              type            = click.STRING,
              default         = "\t",
              required        = False)
def reference_lengths(ribo, out, sep):
    """Gives transcript names and their lengths.

    Annotation is written to standard output if no
    output file is given.

    The first column corresponds to transcript names 
    and the second column corresponds to transcript
    lengths.

    Examples:
     Print the output to the terminal

     .. code:: bash
     
         ribog dump reference-lengths sample.ribo

     Save the output, in gzipped form, in lengths.csv.gz
     and use , to separate columns

     .. code:: bash

         ribog dump reference-lengths -o lengths.csv.gz --sep "," sample.ribo
    """
    return dump_reference_lengths_wrapper(ribo, out, sep = sep)

#########################################################

@dump.command()
@click.argument("ribo",    type = click.Path(exists=True))
@click.argument("experiment", type = click.STRING)
@click.option( '-o', '--out',
               help     = 'Output file in csv format')
@click.option( '--lowerlength',
               type     = click.INT,
               help     = 'Minimum read length to take')
@click.option( '--upperlength',
               type     = click.INT,
               help     = 'Maximum read length to take')
@click.option( '--format',
               type     = click.Choice(['bg', 'tsv']),
               help     = 'Output file format')
def coverage(ribo, 
             out, 
             experiment,
             lowerlength, upperlength, 
             format):
   """
   Prints the coverage of each transcript.

   This command prints the coverage of each transcript, 
   at nucleotide resolution, for a given read length range.
   For a single read length, set 
   'lowerlength' equal to 'upperlength'.
   The values for the given range are aggregated by summation.

   Append ".gz" to the output file name to have the 
   output in zipped form. 

   File format:     
    options: bg, tsv

    bg: Bedgraph
     The columns of the bedgraph file are of the form
     transcript_name location_start location_end coverage_value
     Only the nonzero values are reported in the bedgraph file.
     Bedgraph is 0-based and location_start is inclusive and
     location_end is exclusive. 

     See https://genome.ucsc.edu/goldenPath/help/bedgraph.html
     for a detailed description of bedgraph file format.

    tsv: Tab separated File
      In this format, coverage values for all nucleotide positions,
      regardless of their value ( 0 or not ) are reported.

      The columns of the file are of the form
      transcript_name comma_separated_coverage 

   \b
    
   Examples:  
    1) Get coverage data, of WT, for read length 21 in zipped bedgraph format

    \b
    
    .. code:: bash 
       
               ribog --lowerlength 21 --upperlength 21 \\
                   -o coverage.bg.gz \\
                   --fromat bg sample.ribo WT


    2) Get coverage data, of treatment_1, for range 26 to 30 in tsv format

    \b
    
    .. code:: bash 
       
           ribog --lowerlength 26 --upperlength 30 
                 -o coverage.tsv  
                 --format tsv 
                 sample.ribo treatment_1
   """

   dump_coverage_wrapper(ribo_file       = ribo, 
                         output_file     = out, 
                         experiment_name = experiment,
                         range_lower     = lowerlength, 
                         range_upper     = upperlength,
                         file_format     = format)



# -*- coding: utf-8 -*-

###########################################################

def _print_coverage_bg(cov_dict, file_handle):
    """
    Print cov_dict in bedgraph format.
    """

    for transcript, t_cov in cov_dict.items():
        for i, i_cov in enumerate(t_cov):
            if i_cov == 0:
                continue
            line = "\t".join( 
                         tuple( 
                            map( str, (transcript, 
                                       i, 
                                       i + 1, 
                                       i_cov)  )  ) )
            print(line, file = file_handle)


def _print_coverage_tsv(cov_dict, file_handle, 
                                   sep = ","):
    """
    Print cov_dict in tab separated file.

    The first column is the transcript name and the
    second column is the coverage values separated by commas.
    """

    for transcript, t_cov in cov_dict.items():
        
        line = transcript + "\t" + ",".\
                  join(tuple( map(str, t_cov) ))                              
        print(line, file = file_handle)


def print_coverage(  cov_dict, 
                     file_handle, 
                     file_format = "bg", 
                     sep         = ","  ):
    """
    Prints the given transcript coverages to a given file handle

    The keys of the coverage dictionary (cov_dict)
    are transcript names. 
    The values are numpy arrays giving the coverage of the transcript, 
    """

    if file_format == "bg":
        _print_coverage_bg( cov_dict    = cov_dict, 
                            file_handle = file_handle)
    else:
        _print_coverage_tsv(cov_dict    = cov_dict, 
                            file_handle = file_handle, 
                            sep         = sep)



# -*- coding: utf-8 -*-
import unittest
import os
from io import StringIO, BytesIO

import numpy as np
import h5py

from test_data import *
from ribogadgets import create
from ribogadgets.core.coverage import find_coverage
from ribogadgets.core.get_gadgets import get_reference_names,\
                                         get_reference_lengths,\
                                         get_region_boundaries
from ribogadgets.settings import *
from ribogadgets.core.quantify import quantify_experiment

###########################################

TRANSCRIPT_LENGTHS=\
"""GAPDH    10
VEGFA   5
FLT 22
MYC 1"""


TRANSCRIPT_ANNOTATION=\
"""GAPDH    0   4  UTR5    0   -
GAPDH   4  7    CDS 0   -
GAPDH   7    10 UTR3   0   -
VEGFA   0   3 CDS 0   +
VEGFA   3 5 UTR3    0   +
FLT   0 6 UTR5    0   +
FLT   6 15 CDS    0   +
FLT   15 20 UTR3    0   +
MYC 0  1    CDS 0   -
MYC 1    1    UTR3    0   -"""

READ_SET_1=\
"""GAPDH 2  4  read_1 0   +
MYC 0  1  read_2 0 +
VEGFA 0  5  read_30 0 +
VEGFA 0  6  read_31 0 +
VEGFA 4  5  read_3 0 +
MYC 0  1  read_4 0 +
MYC 0  1  read_5 0 +
VEGFA 4  5  read_6 0 +
VEGFA 1  2  read_7 0 +
FLT 4  10  read_8 0 +
FLT 4  6  read_9 0 +
FLT 4  8  read_10 0 +
FLT 4  7  read_11 0 +
FLT 4  9  read_12 0 +
FLT 4  5  read_13 0 +
FLT 5  6  read_14 0 +
FLT 5  6  read_15 0 +
FLT 5  7  read_16 0 +
FLT 5  8  read_17 0 +
FLT 6  9  read_18 0 +
FLT 7  8  read_19 0 +
FLT 7  9  read_20 0 +
FLT 7  10  read_21 0 +
FLT 7  11  read_22 0 +
FLT 7  12  read_23 0 +
FLT 8  9  read_24 0 +
FLT 8  11  read_240 0 +
FLT 9  19  read_25 0 +
FLT 10  12  read_26 0 +
FLT 11  15  read_27 0 +
FLT 12  14  read_28 0 +
FLT 13  15  read_29 0 +
FLT 13  16  read_30 0 +
FLT 13  17  read_31 0 +
FLT 13  17  read_32 0 +
FLT 14  17  read_33 0 +
FLT 14  17  read_34 0 +
FLT 16  17  read_35 0 +
FLT 16  17  read_36 0 +
FLT 16  17  read_37 0 +
FLT 17  19  read_38 0 +"""

"""
Expected start site distribution, for radius=2:

GAPDH: [1,0,0,0,0]
VEGFA: [0, 0, 2, 1, 0]
FLT1: [6, 4, 1, 5, 2]
MYC: [0, 0, 3, 0, 0]
"""

"""
Expected stop site distribution, for radius = 2:

GAPDH: [0, 0, 0, 0, 0]
VEGFA: [1, 0, 0, 2, 0]
MYC: [0, 3, 0, 0, 0 ]
FLT1: [4, 2, 0, 3, 1]
"""

ACTUAL_START_SITE_COVERAGE = np.array( [[1, 0, 0, 0, 0], 
                                [0, 0, 2, 1, 0],
                                [6, 4, 1, 5, 2],
                                [0, 0, 3, 0, 0]] )

ACTUAL_STOP_SITE_COVERAGE = np.array( [ [0, 0, 0, 0, 0],
                                        [1, 0, 0, 2, 0],
                                        [4, 2, 0, 3, 1],
                                        [0, 3, 0, 0, 0]] )

###########################################
###########################################

TRANSCRIPT_LENGTHS_2=\
"""GAPDH    90
VEGFA   12
FLT 40
MYC 1
P53 85"""

READ_SET_2=\
"""GAPDH 2  4  read_UTR5_1 0   +
GAPDH 3  5  read_UTR5_2 0   +
GAPDH 10  14  read_UTR5_3 0   +
GAPDH 10  14  read_UTR5_4 0   +
GAPDH 11  14  read_UTR5_5 0   +
GAPDH 12  14  read_UTR5_6 0   +
GAPDH 14  25  read_UTR5_7 0   +
GAPDH 15  20  read_UTR5_junc_1 0   +
GAPDH 16  20  read_UTR5_junc_2 0   +
GAPDH 17  20  read_UTR5_junc_3 0   +
GAPDH 20  25  read_UTR5_junc_4 0   +
GAPDH 23  25  read_UTR5_junc_5 0   +
GAPDH 24  30  read_CDS_1 0   +
GAPDH 30  35  read_CDS_2 0   +
GAPDH 44  46  read_CDS_3 0   +
GAPDH 45  55  read_UTR3_junc_1 0   +
GAPDH 53  57  read_UTR3_junc_2 0   +
GAPDH 54  57  read_UTR3_1 0   +
VEGFA 0  1  read_UTR5_junc_1 0  +
VEGFA 3  5  read_UTR5_junc_2 0  +
VEGFA 4  11  read_CDS_1 0  +
VEGFA 4  5  read_CDS_2 0  +
VEGFA 11  12  read_UTR3_junc_1 0  +
FLT 36  40  read_UTR3_junc_1 0 +
MYC 0  1  read_UTR5_junc_1 0  +
P53 5  15  read_UTR5_1 0  +
P53 25  35  read_CDS_1 0  +"""

ANNOTATION_2 = [\
               ([0, 20], [20, 50], [50, 90]),\
               ([0, 0], [0, 10], [10, 12]),\
               ([0, 15], [15, 40], [40, 40]),\
               ([0, 0], [0, 1], [1, 1]),\
               ([0, 20], [20, 55], [55, 85])
              ]

GAPDH_counts = ( 7, 5, 3, 2, 1 )
VEGFA_counts = ( 0, 2, 2, 1, 0 )
FLT_counts   = ( 0, 0, 0, 1, 0 )
MYC_counts   = ( 0, 1, 0, 0, 0 )
P53_counts   = ( 1, 0, 1, 0, 0 )

expected_counts = ( GAPDH_counts, VEGFA_counts, FLT_counts,
                    MYC_counts, P53_counts )

REF_LEN_FILE_2    = "ref_len_2.txt"
ANNOTATION_FILE_2 = "annotation_2.bed"
ALIGNMENT_FILE_2  = "alignment_2.bed"

###########################################
###########################################


LEFT_SPAN  = 5
RIGHT_SPAN = 3

ANNOTATION = [\
               ([0, 4], [4, 7], [7, 10]),\
               ([0, 0], [0, 3], [3, 5]),\
               ([0, 6], [6, 15], [15, 20]),\
               ([0,0], [0, 1], [1,1])
              ]


def _get_transcripts(file_in_string):
    rows        = file_in_string.split("\n")
    pairs       = tuple( map( lambda x: x.split(), rows ) )
    ref_names   = tuple( map( lambda x: x[0], pairs ) )
    ref_lengths = tuple( map( lambda x: x[1], pairs ) )
    ref_lengths = tuple( map( int, ref_lengths ) )
    return (ref_names, ref_lengths)


class TestCreate(unittest.TestCase):

    def setUp(self):
        self.tmp_files         = list()
        self.ref_len_file      = StringIO(TRANSCRIPT_LENGTHS)
        self.alignment_file_1  = StringIO(READ_SET_1)
        self.ref_len_file_2    = StringIO(TRANSCRIPT_LENGTHS_2)
        self.alignment_file_2  = StringIO(READ_SET_2)


    def tearDown(self):
        pass


    def test_quantify_metagene(self):
        ref_names , ref_lengths = _get_transcripts(TRANSCRIPT_LENGTHS)

        ribo_profile = quantify_experiment( 
                          input_reads_file   = self.alignment_file_1, 
                          ref_names          = ref_names, 
                          ref_lengths        = ref_lengths, 
                          region_coordinates = ANNOTATION, 
                          metagene_radius    = 2, 
                          left_span          = LEFT_SPAN, 
                          right_span         = RIGHT_SPAN )

        start_site_coverage = ribo_profile[REF_DG_START_SITE_COV]
        comparison          = (ACTUAL_START_SITE_COVERAGE == \
                               start_site_coverage)
        self.assertTrue( np.all( comparison ) )

        stop_site_coverage = ribo_profile[REF_DG_STOP_SITE_COV]
        comparison         = (ACTUAL_STOP_SITE_COVERAGE ==\
                              stop_site_coverage)
        self.assertTrue( np.all( comparison ) )

    def test_quantify_regions(self):
        ref_names , ref_lengths = _get_transcripts(TRANSCRIPT_LENGTHS_2)

        ribo_profile = quantify_experiment( 
                          input_reads_file   = self.alignment_file_2, 
                          ref_names          = ref_names, 
                          ref_lengths        = ref_lengths, 
                          region_coordinates = ANNOTATION_2, 
                          metagene_radius    = 2, 
                          left_span          = LEFT_SPAN, 
                          right_span         = RIGHT_SPAN )

        rg = ribo_profile[REF_DG_REGION_COUNTS]

        for i in range(5):
            comparison = (rg[i,: ] == expected_counts[i] )
            self.assertTrue(np.all(comparison) ) 
  

if __name__ == '__main__':
        
    unittest.main()


# -*- coding: utf-8 -*-

from functools   import partial
from collections import OrderedDict
from functools   import partial, wraps

import h5py
import pandas as pd

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
from matplotlib.backends.backend_pdf import PdfPages
# The "use" method is necessary for not seeing 
# gtk error message on the terminal
matplotlib.use('Agg')

from .settings                 import *
from .core.get_gadgets         import *
from .core.aggregate_by_length import aggregate_region_counts,\
                                      aggregate_start_site,\
                                      aggregate_stop_site

from .dump        import *
from .core.verify import verify_extended_region_name,\
                         verify_site_type

from .core.exceptions import MissingExperiment

from .core.verify import make_cli_function,\
                         check_experiment_list_in_ribo_handle,\
                         check_length_range

#####################################################################

def save_plot(plot_func):
    @wraps(plot_func)
    def out_plot_func(*args, **kwargs):
        fig = plt.figure( )
        res = plot_func(*args, **kwargs)
        if "output_file" in kwargs.keys() and kwargs["output_file"]:
            plt.savefig(kwargs["output_file"], bbox_inches = "tight")
        plt.close(fig)
        return res
    return out_plot_func

def fix_experiment_list(func):
    @wraps(func)
    def exp_fixed_func(*args, **kwargs):
        if not kwargs["experiment_list"]:
            raise MissingExperiment("At least one experiment is required!")
        if type(kwargs["experiment_list"]) == str:
            kwargs["experiment_list"] = [kwargs["experiment_list"]]
        return func(*args, **kwargs)
    return exp_fixed_func


####################################################################    

@check_length_range
@fix_experiment_list
@save_plot
@check_experiment_list_in_ribo_handle
def plot_metagene( ribo_handle, 
                   site_type,
                   *, 
                   experiment_list, 
                   title,
                   range_lower = 0, 
                   range_upper = 0,
                   normalize   = False,
                   output_file = "",
                   colors      = PLOT_COLORS ):
    
    """
    Generates metagene plots at start or stop sites.
    If a file name / path is given, it saves the output to the
    provided file with the given name / path.

    At least one experiment must be provided.
    Unless you provide a custom list of colors,
    at most 7 experiments can be drawn on the same plot
    as there are 7 colors available by default.

    normalize = True divides the coverage vector by the sum of the
    vector (not the whole transcript coverage) and multiplies
    by 100 and gives percentage.
    This is especially useful when plotting experiments together.

    All read lengths in the interval range_lower
    and range_upper are aggregated for the resulting plot.



    Parameters
    ----------
    ribo_handle: h5py.File
       ribo file handle

    site_type: str
       for start site, use "start_site_coverage"
       for stop site use "stop_site_coverage"

    experiment_list: list 
       list of experiment names

    title: str
       Title of the graph

    range_lower: int
       Minimum read length to be taken
       If 0, minimum length of the ribo file is taken

    range_upper: int
       Maximum read length to be taken
       If 0, maximum length of the ribo file is taken

    normalize: Boolean
       If false, plot raw counts
       If True, divide the coverage vector 

    output_file: str
       The path of the output file.
       If it is an NOT empty string,
       the plot will be saved in this file

    colors: list(str)
       A list of colors of mathplotlib

    """
            
    if len(experiment_list) > len(colors):
        error_msg =  "There can be at most {} plots.".format(len(colors))
        error_msg += "{} provided.".format(len(experiment_list))
        raise ValueError(error_msg)
    
    verify_libnames_and_readlength_range( ribo_handle      = ribo_handle, 
                                          experiment_list = experiment_list,
                                          range_lower      = range_lower, 
                                          range_upper      = range_upper)

    metagene_data = dump_metagene( ribo_handle     = ribo_handle, 
                                   site_type       = site_type,                           
                                   range_lower     = range_lower, 
                                   range_upper     = range_upper,
                                   sum_lengths     = True, 
                                   sum_references  = True,
                                   experiment_list = experiment_list )    
    
    metagene_vectors = [ metagene_data[lib].iloc[0] for lib in experiment_list ] 
    ylabel = "Frequency Raw"
    
    if normalize:
        ylabel = "Frequency %"
        metagene_vectors = [ (this_vector / sum(this_vector) ) * 100 \
                             for this_vector in metagene_vectors]
    
    metagene_radius = get_metagene_radius(ribo_handle)
    plt.xticks(np.arange(-1*metagene_radius, metagene_radius+1, step=10 ) )

    for i, mg_vector in enumerate(metagene_vectors):
        obj1 = plt.plot(mg_vector,
               linestyle='-', color = colors[i] , linewidth = 1)
    
    plt.title(title)
    plt.xlabel("Relative Position")
    plt.ylabel(ylabel)
    if len(experiment_list) > 1:
        plt.legend(experiment_list, loc = (1.04,0))

    return metagene_data
    


@make_cli_function
def plot_metagene_wrapper( 
                   ribo_file, 
                   site_type, 
                   range_lower, range_upper,
                   experiment_list, 
                   output_file, 
                   title,
                   normalize    = False,
                   dump_to_file = "" ):
    """
    This is a wrapper function for plot_metagene
    See its documentation for details.

    This function is typically called by CLI
    and calls the actual plot function plot_metagene

    If dump_to_file is a non-empty string,
    the data used for the metagene plot is written
    in the provided file.
    """

    if not experiment_list:
        exit("Please provide experiment names.")


    with h5py.File(ribo_file, "r") as ribo_handle:
        if site_type.lower() == "start":
            site_type_actual = REF_DG_START_SITE_COV
        elif site_type.lower() == "stop":
            site_type_actual = REF_DG_STOP_SITE_COV
        else:
            raise ValueError("Wrong site type:", site_type)
        
        min_length, max_length = get_read_length_range(ribo_handle)

        plot_metagene( ribo_handle      = ribo_handle, 
                       site_type        = site_type_actual, 
                       range_lower      = range_lower, 
                       range_upper      = range_upper,
                       experiment_list = experiment_list, 
                       title            = title,
                       normalize        = normalize,
                       output_file      = output_file )
        

    if dump_to_file:
        dump_metagene_wrapper( ribo_file       = ribo_file, 
                               output_file     = dump_to_file, 
                               site_type       = site_type, 
                               sum_lengths     = True, 
                               sum_references  = True,
                               range_lower     = range_lower, 
                               range_upper     = range_upper,
                               experiment_list = experiment_list )

            
@fix_experiment_list
@save_plot
@check_experiment_list_in_ribo_handle
def plot_lengthdist( ribo_handle,
                     *, 
                     region_type, 
                     experiment_list, 
                     title, 
                     normalize   = False,
                     output_file = "",
                     colors      = PLOT_COLORS ):
    
    """
    Plots the distribution of the lengths of ribosome 
    footprints on a given region.

    If normalize is set to True, the values are divided
    by the sum and multiplied by 100 to give 
    the percentage on the y-axis.

    The number of experiments can not exceed the number
    of the elements os colors argument.

    Output file can be a pdf or a png file,

    Parameters
    ----------

    ribo_handle: h5py.File
      Open h5py handle

    region_type: str
      Transcript region to be plotted.
      It can be one of the UTR5, CDS, UTR3

    experiment_list: list(str)
      Name of the experiments to be plotted.

    title: str
      Title of the plot to appear at the top

    normalize: Boolean
       Determines whether raw frequencies 
       or percentages are to be plotted.

    output_file: str
       If provided, the resulting plot will be saved
       in this file. 
       pdf and png extensions are supported.

    colors: list(str)
       Colors of the individual experiments 
       to be passed on to matplotlib.

    """

    min_length, max_length = get_read_length_range(ribo_handle)

    region_data = dump_region(
                    ribo_handle     = ribo_handle, 
                    region_name     = region_type,
                    sum_lengths     = False, 
                    sum_references  = True, 
                    range_lower     = min_length , 
                    range_upper     = max_length ,
                    experiment_list = experiment_list)
    
    ylabel = "Frequency"
    
    if normalize:
        ylabel = "Frequency %"
        for name in experiment_list:
            length_dist       = np.array(region_data[name], dtype = np.int64)
            length_sum        = np.sum(length_dist)
            region_data[name] = ( length_dist / length_sum ) * 100

    plt.xticks(np.arange(min_length, max_length+1, step=2 ) )

    for i, name in enumerate(experiment_list):
        obj1 = plt.plot(region_data[name],
                   linestyle='-', color = colors[i] , linewidth = 1)
        
    if len(experiment_list) > 1:
        plt.legend(experiment_list, loc=(1.04,0))
    
    plt.title(title)
    plt.xlabel("Read Length")
    plt.ylabel(ylabel)    

    return region_data


@make_cli_function
def plot_lengthdist_wrapper( 
                   ribo_file, 
                   region_type, 
                   experiment_list, 
                   output_file, 
                   title,
                   normalize,
                   dump_to_file = "" ):

    """
    This is a wrapper function for plot_lengthdist
    See its documentation for details.

    This function is typically called by CLI
    and calls the actual plot function plot_lengthdist

    If dump_to_file is a non-empty string,
    the data used for the metagene plot is written
    in the provided file.
    """
    
    with h5py.File(ribo_file, "r") as ribo_handle:
        min_length, max_length = get_read_length_range(ribo_handle)

        plot_lengthdist( ribo_handle      = ribo_handle, 
                         region_type      = region_type, 
                         experiment_list = experiment_list, 
                         title            = title, 
                         normalize        = normalize,
                         output_file      = output_file)


    if dump_to_file:
        dump_region_wrapper(ribo_file       = ribo_file, 
                            output_file     = dump_to_file, 
                            region_name     = region_type,
                            sum_lengths     = False, 
                            sum_references  = True, 
                            range_lower     = min_length, 
                            range_upper     = max_length,
                            experiment_list = experiment_list)

@check_length_range
@fix_experiment_list
@save_plot
@check_experiment_list_in_ribo_handle
def plot_region_counts(ribo_handle, *,
                       experiment_list = [],
                       title            = "",
                       range_lower      = 0, 
                       range_upper      = 0, 
                       output_file      = ""):
    """
    Generates barplots for the percentage of the reads
    of the regions UTR5, CDS and UTR3 for the provided
    experiments.

    Parameters
    ----------
    ribo_handle: h5py.File
      Open h5py handle

    experiment_list: list(str)
      Name of the experiments to be plotted.

    title: str
      Title of the plot to appear at the top

    output_file: str
       If provided, the resulting plot will be saved
       in this file. 
       pdf and png extensions are supported.

    Returns
    -------
    pd.DataFrame: Dataframe used to generate the plot.
    """
    
    region_getter = partial( dump_region, 
                             ribo_handle     = ribo_handle, 
                             sum_lengths     = True, 
                             sum_references  = True, 
                             range_lower     = range_lower, 
                             range_upper     = range_upper,
                             experiment_list = experiment_list)
    
    region_list   = (UTR5_name, CDS_name, UTR3_name)
    region_counts = [ region_getter(region_name = r) for r in region_list ]
    
    for i, region in enumerate(region_list):
        region_counts[i].rename({region_counts[i].columns[0]:region} ,
                                axis="columns", inplace=True)
        
    region_df     = pd.concat(region_counts, axis = 1, join = "outer") 
    percentage_df = region_df.divide(
                      other = region_df.sum(axis=1), axis = 0 ) \
                                             * 100 
    
    ax  = percentage_df.plot.bar(stacked=True)
    plt.title(title)
    plt.legend(loc = (1.04,0))
    
    alignment       = {'horizontalalignment': 'center', 
                       'verticalalignment'  :   'baseline'}
    cds_percentages = list(map("{0:.1f}".format,percentage_df[CDS_name]))
    
    for i, p in enumerate(cds_percentages):
        plt.text(i, 50, p, **alignment)
    
    return region_df


@make_cli_function
def plot_region_counts_wrapper(ribo_file, 
                               experiment_list, 
                               title,
                               range_lower, range_upper, 
                               output_file, 
                               dump_to_file):

    """
    A wrapper function for plot_region_counts

    This function is typically called by CLI.
    """

    with h5py.File(ribo_file, "r") as ribo_handle:
        count_df = plot_region_counts(
                           ribo_handle, 
                           experiment_list = experiment_list,
                           title            = title,
                           range_lower      = range_lower, 
                           range_upper      = range_upper, 
                           output_file      = output_file)

    if dump_to_file:
        count_df.to_csv(dump_to_file, sep = DF_CSV_SEPARATOR)


# -*- coding: utf-8 -*-
from functools import partial

from .get_gadgets import *
from ..settings import *

###########################################################

def sum_individual_dataset_list(dataset, 
                                number_of_references, 
                                lower_relative_index, 
                                upper_relative_index):
    """
    Sums the counts coming from an individual dataset
    of one experiment.

    For each reference (transcript), it grabs the counts in the
    given range and sums them and reports the sum.
    These inintial counts, being summed, are typically coming 
    from the hdf5 dataset directly.

    Both range values are inclusive.

    Parameters
    ----------

    dataset: Actual dataset of count values from hdf5

    number_of_references: number of references (transcripts)
                          in the annotation of the ribo file

    lower_relative_index: relative position of the initial
                          length relative to the minimum 
                          read length of the ribo file
                          i.e.: length_range_min - ribo_length_min

    upper_relative_index: relative position of the terminal
                          length relative to the maximum 
                          read length of the ribo file
                          i.e.: lower_relative_index + \
                                (range_max - range_min)

    Returns
    -------
    result: 2D numpy array of dimension 
            ( number_of_transcripts(references), ncols_of_dataset )

    """
    
    if len(dataset.shape) == 2:
        ncols  = dataset.shape[1]
        result = np.zeros( (number_of_references, ncols), 
                           dtype = DEFAULT_COUNT_DT )
    else:
        result = np.zeros( number_of_references, 
                           dtype = DEFAULT_COUNT_DT )

    for i in range(lower_relative_index, upper_relative_index + 1):
        result += dataset[i*number_of_references:\
                          (i+1)*number_of_references]

    return result

############################################################    

def group_individual_dataset_list(dataset, 
                                  number_of_references,
                                  range_lower,
                                  range_upper,
                                  lower_relative_index,  
                                  upper_relative_index):
    """
    Grabs a slice of the dataset for a given relative index range
    Adds the corresponding read lengths in the first column.
    In contrast to the sum_individual_dataset_list function,
    it does not add count values up.


    Parameters
    ----------

    dataset: Actual dataset of count values from hdf5

    number_of_references: number of references (transcripts)
                          in the annotation of the ribo file

    range_lower: actual minimum read length to be taken
                 This is the initial value of the first column
                 of the resulting np array

    range_upper: actual maximum read length to be taken
                 This is the terminal value of the first column
                 of the resulting np array

    lower_relative_index: relative position of the initial
                          length relative to the minimum 
                          read length of the ribo file
                          i.e.: length_range_min - ribo_length_min

    upper_relative_index: relative position of the terminal
                          length relative to the maximum 
                          read length of the ribo file
                          i.e.: lower_relative_index + \
                                (range_max - range_min)

    Returns
    -------
    A numpy array of the following  form where
    the first column corresponds to the read lengths
    that the counts are coming from
    [ range_lower, count_values_of_length_range_lower ]
    [ range_lower+1, count_values_of_length_range_lower+1 ]
    [ ..........., count_values_.... ]
    [ range_upper, count_values_of_length_range_upper ]

    """
    
    length_column = np.repeat( range(range_lower, range_upper + 1), 
                               number_of_references )
    length_column = np.array( length_column , ndmin = 2, 
                              dtype = DEFAULT_COUNT_DT )

    data_in_range_of_interest = \
        dataset[ lower_relative_index * number_of_references :\
                 (upper_relative_index + 1) * number_of_references ]


    return np.concatenate( (length_column.T, 
                               data_in_range_of_interest), 
                                 axis = 1 )

#######################################################

def aggregate_by_length( ribo_handle, 
                         dataset_name,
                         range_lower, 
                         range_upper,
                         sum_values      = False, 
                         experiment_list = []  ):

    """
    Aggregates values of a dataset by grouping by length 
    or by summing them up for a given read length range.

    This is done for every experiment in the given experiment
    in the experiment_list.
    If the list is empty, then all experiments in ribo
    file handle are taken.

    Parameters
    ----------
    ribo_handle : HDF5 handle of the open ribo file

    dataset_name : Used as a path for the dataset of interest
                   in the ribo file

    range_lower : initial read length to be taken

    range_upper : initial read length to be taken

    sum_values : If true, values are summed up
                 in the given read length range.
                 If false, all the summands are reported
                 with their corresponding read length

    experiment_list : The experiments to be reported.
                  If the list is empty, all experiments in the ribo file are to be taken

    Returns
    -------
    A 2D Numpy Array consisting of Aggregated data by length

    """
    
    if experiment_list != []:
        actual_experiment_list = experiment_list
    else:
        actual_experiment_list = get_experiment_names( ribo_handle )

    dataset_list = tuple(map( lambda x: \
                          ribo_handle[ EXPERIMENTS_name ][x][dataset_name][...], 
                          actual_experiment_list ) )

    read_len_min, read_len_max = get_read_length_range(ribo_handle)

    lower_relative_index = range_lower - read_len_min
    if lower_relative_index < 0:
        raise Exception("Lower range ({}) can not be".format(range_lower) + \
                        " smaller than minimum" +\
                        " read length ({})".format(read_len_min))

    upper_relative_index = (range_upper - range_lower) + lower_relative_index  
    if range_upper < range_lower or range_upper > read_len_max:
        raise Exception("Upper range must be less than the maximum"
                        " read length of the ribo and" 
                        "greater than the lower range")
    
    number_of_references = get_number_of_references(ribo_handle)

    if sum_values:
        aggregate_func = partial(sum_individual_dataset_list,
                          number_of_references = number_of_references, 
                          lower_relative_index = lower_relative_index, 
                          upper_relative_index = upper_relative_index)
    else:
        aggregate_func = partial(group_individual_dataset_list, 
                          number_of_references = number_of_references, 
                          range_lower          = range_lower,
                          range_upper          = range_upper,
                          lower_relative_index = lower_relative_index,  
                          upper_relative_index = upper_relative_index)

    aggregated_data = tuple( map(aggregate_func, dataset_list ) )

    return aggregated_data

############################################################

def aggregate_start_site(ribo_handle, range_lower, range_upper,
                         sum_values      = False,
                         experiment_list = []):
    """
    A wrapper function for aggregate_by_length for start sites
    See the documentation of aggregate_by_length for details
    """

    dataset_name = METAGENE_name + "/" + REF_DG_START_SITE_COV
    return aggregate_by_length(ribo_handle, dataset_name,
                         range_lower, range_upper,
                         sum_values, experiment_list)

def aggregate_stop_site(ribo_handle, range_lower, range_upper, 
                        sum_values      = False,
                        experiment_list = []):
    """
    A wrapper function for aggregate_by_length for stop sites
    See the documentation of aggregate_by_length for details
    """

    dataset_name = METAGENE_name + "/" + REF_DG_STOP_SITE_COV
    return aggregate_by_length(ribo_handle, dataset_name,
                         range_lower, range_upper, 
                         sum_values, experiment_list)

def aggregate_region_counts( ribo_handle, range_lower, range_upper, 
                             sum_values      = False,
                             experiment_list = [] ):
    """
    A wrapper function for aggregate_by_length for region counts
    See the documentation of aggregate_by_length for details
    """

    dataset_name = REF_DG_REGION_COUNTS + "/" + REF_DG_REGION_COUNTS
    return aggregate_by_length(ribo_handle, dataset_name,
                         range_lower, range_upper, 
                         sum_values, experiment_list)



# -*- coding: utf-8 -*-
import click

@click.group()
@click.option('--debug/--no-debug', default = False)
def cli(debug):
    pass
    
from .create   import *
from .merge    import *
from .info     import *
from .dump     import *
from .plot     import *
from .metadata import *
from .rnaseq    import *



if __name__ == "__main__":

    cli()


